#############################################################################################
#                                                                                           #
# Author:   GeoPy Team                                                                      #
# Email:    geopy.info@gmail.com                                                            #
# Date:     March 2018                                                                      #
#                                                                                           #
#############################################################################################

# seismic data IO
from PyQt5 import QtCore
import os
import sys
import numpy as np
import math
#
sys.path.append(os.path.dirname(__file__)[:-8][:-4][:-13])
from cognitivegeo.src.vis.messager import messager as vis_msg
from cognitivegeo.src.seismic.analysis import analysis as seis_ays
from cognitivegeo.src.segpy.header import field
from cognitivegeo.src.segpy.types import Int32
import cognitivegeo.src.segpy.reader as syreader
import cognitivegeo.src.segpy.writer as sywriter
import cognitivegeo.src.segpy.packer as sypacker
from cognitivegeo.src.segpy.binary_reel_header import BinaryReelHeader
from cognitivegeo.src.segpy.trace_header import TraceHeaderRev1

__all__ = ['inputoutput']


def readSeisDataFromSegy(segyfile, seisinfo, traceheaderformat=TraceHeaderRev1, endian='>',
                         verbose=True, qpgsdlg=None):
    """
    Read seismic data from segy file (by segpy)

    Args:
        segyfile:           segy file for reading
        seisinfo:           seismic information dictionary
        traceheaderformat:  Trace header format. Default is TraceHeaderRev1
        endian:             Endian ('>' or '<'). Default is '>'
        verbose:            Verbose option
        qpgsdlg:            Qt preogress dialog

    Return:
        seisdata:   Seismic data in 3D matrix
        traceflag:  Flag for seismic traces. 1 for missing traces
    """

    inlstart = seisinfo['ILStart']
    inlstep = seisinfo['ILStep']
    inlnum = seisinfo['ILNum']
    xlstart = seisinfo['XLStart']
    xlstep = seisinfo['XLStep']
    xlnum = seisinfo['XLNum']
    znum = seisinfo['ZNum']
    #
    seisdata = np.zeros([znum, xlnum, inlnum])
    traceflag = np.zeros([xlnum, inlnum])
    #
    traceheaderpacker = sypacker.HeaderPacker(traceheaderformat, endian)
    dataformat = np.array([4, 4, 2, 4, 4, 0, 0, 1])
    #
    with open(segyfile, 'rb') as sid:
        dat = syreader.create_reader(sid, trace_header_format=traceheaderformat, endian=endian)
        binary_header = syreader.read_binary_reel_header(sid, endian=endian)
        databyte = dataformat[binary_header.data_sample_format - 1]
        #
        pos = np.int64(3200 + 400)
        if qpgsdlg is not None:
            qpgsdlg.setMaximum(inlnum*xlnum)
        for itrace in dat.trace_indexes():
            #
            if qpgsdlg is not None:
                QtCore.QCoreApplication.instance().processEvents()
                qpgsdlg.setValue(itrace)
            #
            if verbose:
                sys.stdout.write(
                    '\r>>> Read %d traces: %.1f%% ' % (inlnum*xlnum, float(itrace)/float(inlnum*xlnum)*100))
                sys.stdout.flush()
            #
            header = syreader.read_trace_header(sid, trace_header_packer=traceheaderpacker, pos=pos)
            pos = pos + 240 + header.num_samples * databyte
            #
            idx_xl = int((header.crossline_number - xlstart) / xlstep)
            idx_inl = int((header.inline_number - inlstart) / inlstep)
            idx = idx_xl + idx_inl * xlnum
            if idx < (inlnum * xlnum):
                seisdata[:, idx_xl, idx_inl] = dat.trace_samples(itrace)
                traceflag[idx_xl, idx_inl] = 0
        #
        if qpgsdlg is not None:
            qpgsdlg.setValue(inlnum*xlnum)
        if verbose:
            print('Done')
    #
    return seisdata, traceflag


def readSeisInfoFromSegy(segyfile, traceheaderformat=TraceHeaderRev1, endian='>'):
    """
    Read basic information of a 3D seismic survey from a segy file

    Args:
        segyfile:           file name of seismic file in segy format
        traceheaderformat:  The class defining the layout of the trace header.
        endian:             '>' for big-endian data (the standard and default), '<'
                                for little-endian (non-standard)
    Returns:
        Seismic information dictionary as defined in analysis.py

    Note:
        negative z is used in the vertical direction
    """

    # Check if file exist
    if os.path.isfile(segyfile) is False:
        vis_msg.print('ERROR in readSeisInfoFromSegy: ' + segyfile + ' not found', type='error')
        sys.exit()

    with open(segyfile, 'rb') as sid:
        dat = syreader.create_reader(sid, trace_header_format=traceheaderformat, endian=endian)
        #
        if (dat.dimensionality > 3 and dat.dimensionality <= 0):
            vis_msg.print('ERROR in readSeisInfoFromSegy: ' + segyfile + ' contains no seismic survey',
                          type='error')
            sys.exit()

        if (dat.dimensionality == 1):
            vis_msg.print('WARNING in readSeisInfoFromSegy: ' + segyfile + ' contains 1D seismic survey',
                          type='warning')
            inlstart = 0
            inlend = 0
            inlnum = 1
            xlstart = 0
            xlend = 0
            xlnum = 1

        if (dat.dimensionality == 2):
            vis_msg.print('WARNING in readSeisInfoFromSegy: ' + segyfile + ' contains 2D seismic survey',
                          type='warning')
            inlstart = 0
            inlend = 0
            inlnum = 1
            xlstart = 0
            xlend = dat.num_traces() - 1
            xlnum = dat.num_traces()

        if (dat.dimensionality == 3):
            inlstart = dat.inline_range()[0]
            inlend = dat.inline_range()[-1]
            inlnum = dat.num_inlines()
            xlstart = dat.xline_range()[0]
            xlend = dat.xline_range()[-1]
            xlnum = dat.num_xlines()

        seisinfo = seis_ays.makeSeisInfo()

        # Inline
        inlstep = 1
        if inlnum > 1:
            inlstep = (inlend - inlstart) / (inlnum - 1)
        inlstart = np.round(inlstart).astype(np.int32)
        inlstep = np.round(inlstep).astype(np.int32)
        inlnum = np.round(inlnum).astype(np.int32)
        if inlstep == 0:
            inlstep = 1
        inlend = (inlstart + (inlnum - 1) * inlstep).astype(np.int32)
        inlrange = np.linspace(inlstart, inlend, inlnum).astype(np.int32)
        # Add to seisinfo
        seisinfo['ILStart'] = inlstart
        seisinfo['ILEnd'] = inlend
        seisinfo['ILStep'] = inlstep
        seisinfo['ILNum'] = inlnum
        seisinfo['ILRange'] = inlrange
        # Crossline
        xlstep = 1
        if xlnum > 1:
            xlstep = (xlend - xlstart) / (xlnum - 1)
        xlstart = np.round(xlstart).astype(np.int32)
        xlstep = np.round(xlstep).astype(np.int32)
        xlnum = np.round(xlnum).astype(np.int32)
        if xlstep == 0:
            xlstep = 1
        xlend = (xlstart + (xlnum - 1) * xlstep).astype(np.int32)
        xlrange = np.linspace(xlstart, xlend, xlnum).astype(np.int32)
        # Add to seisinfo
        seisinfo['XLStart'] = xlstart
        seisinfo['XLEnd'] = xlend
        seisinfo['XLStep'] = xlstep
        seisinfo['XLNum'] = xlnum
        seisinfo['XLRange'] = xlrange
        # Z
        zstart = dat.trace_header(0).lag_time_a
        zstep = - dat.trace_header(0).sample_interval / 1000
        znum = dat.trace_header(0).num_samples
        zstart = np.round(zstart).astype(np.int32)
        zstep = np.round(zstep).astype(np.int32)
        znum = np.round(znum).astype(np.int32)
        if zstart > 0:
            zstart = - zstart
        if zstep == 0:
            zstep = -1
        zend = (zstart + (znum - 1) * zstep).astype(np.int32)
        zrange = np.linspace(zstart, zend, znum).astype(np.int32)
        # Add to seisinfo
        seisinfo['ZStart'] = zstart
        seisinfo['ZEnd'] = zend
        seisinfo['ZStep'] = zstep
        seisinfo['ZNum'] = znum
        seisinfo['ZRange'] = zrange

    return seisinfo


def readSeis2DMatFromSegyWithInfo(segyfile, seisinfo=None, traceheaderformat=TraceHeaderRev1, endian='>',
                                  verbose=True, qpgsdlg=None):
    """
    Read seismic data from a segy file, as a 2D matrix, with seismic information provided

    Args:
        segyfile:           file name of seismic file in segy format
        seisinfo:           Seismic information dictionary. If not provided, a scanning will run on the segy file.
        traceheaderformat:  The class defining the layout of the trace header.
        endian:             '>' for big-endian data (the standard and default), '<'
                                for little-endian (non-standard)

    Returns:
        seis2dmat:  a 2D matrix of four cloumns [IL, XL, Z, Value]

    Note:
        Negative z is used in the vertical direction
    """

    # Check if file exist
    if os.path.isfile(segyfile) == False:
        vis_msg.print('ERROR in readSeis2DMatFromSegyWithInfo: ' + segyfile + ' not found', type='error')
        sys.exit()

    if seisinfo is None or seis_ays.checkSeisInfo(seisinfo) is False:
        seisinfo = readSeisInfoFromSegy(segyfile, traceheaderformat=traceheaderformat, endian=endian)

    seis3dmat, traceflag = readSeis3DMatFromSegyWithInfo(segyfile, seisinfo,
                                                         traceheaderformat=traceheaderformat, endian=endian,
                                                         verbose=verbose, qpgsdlg=qpgsdlg)

    seis2dmat = seis_ays.convertSeis3DMatTo2DMat(seis3dmat, seisinfo)

    return seis2dmat


def readSeis2DMatFromSegyNoInfo(segyfile, traceheaderformat=TraceHeaderRev1, endian='>',
                                verbose=True, qpgsdlg=None):
    """
    Read seismic data from a segy file, as a 2D matrix, without seismic information provided

    Args:
        segyfile:           File name of seismic file in segy format
        traceheaderformat:  The class defining the layout of the trace header.
        endian:             '>' for big-endian data (the standard and default), '<'
                                for little-endian (non-standard)

    Returns:
        seis2dmat:  a 2D matrix of four cloumns [IL, XL, Z, Value]

    Note:
        Negative z is used in the vertical direction
    """

    # Check if file exist
    if os.path.isfile(segyfile) == False:
        vis_msg.print('ERROR in readSeis2DMatFromSegyNoInfo: ' + segyfile + ' not found', type='error')
        sys.exit()

    seis3dmat, seisinfo = readSeis3DMatFromSegyNoInfo(segyfile,
                                                      traceheaderformat=traceheaderformat, endian=endian,
                                                      verbose=verbose, qpgsdlg=qpgsdlg)

    seis2dmat = seis_ays.convertSeis3DMatTo2DMat(seis3dmat, seisinfo)

    return seis2dmat


def readSeis3DMatFromSegyWithInfo(segyfile, seisinfo=None, traceheaderformat=TraceHeaderRev1, endian='>',
                                  verbose=True, qpgsdlg=None):
    """
    Read seismic data from a segy file, as a 3D matrix, with seismic information provided

    Args:
        segyfile:           File name of seismic file in segy format
        seisinfo:           Seismic information dictionary. If not provided, a scanning will run on the segy file.
        traceheaderformat:  The class defining the layout of the trace header.
        endian:             '>' for big-endian data (the standard and default), '<'
                                for little-endian (non-standard)
        verbose:            Message display. Default is True
        qpgsdlg:            Qt progress dialog

    Returns:
        seis3dmat:  a 3D matrix [Z/Crossline/Inline]

    Note:
        Negative z is used in the vertical direction
    """

    # Check if file exist
    if os.path.isfile(segyfile) == False:
        vis_msg.print('ERROR in readSeis3DMatFromSegy: ' + segyfile + ' not found', type='error')
        sys.exit()

    if seisinfo is None or seis_ays.checkSeisInfo(seisinfo) is False:
        seisinfo = readSeisInfoFromSegy(segyfile, traceheaderformat=traceheaderformat, endian=endian)

    seis3dmat, traceflag = readSeisDataFromSegy(segyfile, seisinfo,
                                                traceheaderformat=traceheaderformat, endian=endian,
                                                verbose=verbose, qpgsdlg=qpgsdlg)

    return seis3dmat


def readSeis3DMatFromSegyNoInfo(segyfile, traceheaderformat=TraceHeaderRev1, endian='>',
                                verbose=True, qpgsdlg=None):
    """
    Read seismic data from a segy file, as a 3D matrix, without seismic information provided

    Args:
        segyfile:           File name of seismic file in segy format
        traceheaderformat:  The class defining the layout of the trace header.
        endian:             '>' for big-endian data (the standard and default), '<'
                                for little-endian (non-standard)
        verbose:            Message display. Default is True
        qpgsdlg:            Qt progress dialog

    Returns:
        seis3dmat:  a 3D matrix [Z/Crossline/Inline]

    Note:
        Negative z is used in the vertical direction
    """

    # Check if file exist
    if os.path.isfile(segyfile) is False:
        vis_msg.print('ERROR in readSeisInfoFromSegy: ' + segyfile + ' not found', type='error')
        sys.exit()

    traceheaderpacker = sypacker.HeaderPacker(traceheaderformat, endian)
    dataformat = np.array([4, 4, 2, 4, 4, 0, 0, 1])

    with open(segyfile, 'rb') as sid:
        dat = syreader.create_reader(sid, trace_header_format=traceheaderformat, endian=endian)
        binary_header = syreader.read_binary_reel_header(sid, endian=endian)
        databyte = dataformat[binary_header.data_sample_format - 1]
        #
        if (dat.dimensionality > 3 and dat.dimensionality <= 0):
            vis_msg.print('ERROR in readSeis3DMatFromSegyNoInfo: ' + segyfile + ' contains no seismic survey',
                          type='error')
            sys.exit()

        if (dat.dimensionality == 1):
            vis_msg.print('WARNING in readSeisInfoFromSegy: ' + segyfile + ' contains 1D seismic survey',
                          type='warning')
            inlstart = 0
            inlend = 0
            inlnum = 1
            xlstart = 0
            xlend = 0
            xlnum = 1

        if (dat.dimensionality == 2):
            vis_msg.print('WARNING in readSeis3DMatFromSegyNoInfo: ' + segyfile + ' contains 2D seismic survey',
                          type='warning')
            inlstart = 0
            inlend = 0
            inlnum = 1
            xlstart = 0
            xlend = dat.num_traces() - 1
            xlnum = dat.num_traces()

        if (dat.dimensionality == 3):
            inlstart = dat.inline_range()[0]
            inlend = dat.inline_range()[-1]
            inlnum = dat.num_inlines()
            xlstart = dat.xline_range()[0]
            xlend = dat.xline_range()[-1]
            xlnum = dat.num_xlines()

        seisinfo = seis_ays.makeSeisInfo()

        # Inline
        inlstep = 1
        if inlnum > 1:
            inlstep = (inlend - inlstart) / (inlnum - 1)
        inlstart = np.round(inlstart).astype(np.int32)
        inlstep = np.round(inlstep).astype(np.int32)
        inlnum = np.round(inlnum).astype(np.int32)
        if inlstep == 0:
            inlstep = 1
        inlend = (inlstart + (inlnum - 1) * inlstep).astype(np.int32)
        inlrange = np.linspace(inlstart, inlend, inlnum).astype(np.int32)
        # Add to seisinfo
        seisinfo['ILStart'] = inlstart
        seisinfo['ILEnd'] = inlend
        seisinfo['ILStep'] = inlstep
        seisinfo['ILNum'] = inlnum
        seisinfo['ILRange'] = inlrange
        # Crossline
        xlstep = 1
        if xlnum > 1:
            xlstep = (xlend - xlstart) / (xlnum - 1)
        xlstart = np.round(xlstart).astype(np.int32)
        xlstep = np.round(xlstep).astype(np.int32)
        xlnum = np.round(xlnum).astype(np.int32)
        if xlstep == 0:
            xlstep = 1
        xlend = (xlstart + (xlnum - 1) * xlstep).astype(np.int32)
        xlrange = np.linspace(xlstart, xlend, xlnum).astype(np.int32)
        # Add to seisinfo
        seisinfo['XLStart'] = xlstart
        seisinfo['XLEnd'] = xlend
        seisinfo['XLStep'] = xlstep
        seisinfo['XLNum'] = xlnum
        seisinfo['XLRange'] = xlrange
        # Z
        zstart = dat.trace_header(0).lag_time_a
        zstep = - dat.trace_header(0).sample_interval / 1000
        znum = dat.trace_header(0).num_samples
        zstart = np.round(zstart).astype(np.int32)
        zstep = np.round(zstep).astype(np.int32)
        znum = np.round(znum).astype(np.int32)
        if zstart > 0:
            zstart = - zstart
        if zstep == 0:
            zstep = -1
        zend = (zstart + (znum - 1) * zstep).astype(np.int32)
        zrange = np.linspace(zstart, zend, znum).astype(np.int32)
        # Add to seisinfo
        seisinfo['ZStart'] = zstart
        seisinfo['ZEnd'] = zend
        seisinfo['ZStep'] = zstep
        seisinfo['ZNum'] = znum
        seisinfo['ZRange'] = zrange
        #
        seisdata = np.zeros([znum, xlnum, inlnum])
        traceflag = np.zeros([xlnum, inlnum])
        #
        pos = np.int64(3200 + 400)
        if qpgsdlg is not None:
            qpgsdlg.setMaximum(inlnum * xlnum)
        for itrace in dat.trace_indexes():
            #
            if qpgsdlg is not None:
                QtCore.QCoreApplication.instance().processEvents()
                qpgsdlg.setValue(itrace)
            #
            if verbose:
                sys.stdout.write(
                    '\r>>> Scan %s: %d traces ' % (segyfile, itrace + 1))
                sys.stdout.flush()
            #
            header = syreader.read_trace_header(sid, trace_header_packer=traceheaderpacker, pos=pos)
            pos = pos + 240 + header.num_samples * databyte
            #
            idx_xl = int((header.crossline_number - xlstart) / xlstep)
            idx_inl = int((header.inline_number - inlstart) / inlstep)
            idx = idx_xl + idx_inl * xlnum
            if idx < (inlnum * xlnum):
                seisdata[:, idx_xl, idx_inl] = dat.trace_samples(itrace)
                traceflag[idx_xl, idx_inl] = 0
        #
        if qpgsdlg is not None:
            qpgsdlg.setValue(inlnum * xlnum)
        if verbose:
            print('Done')

    return seisdata, seisinfo


def writeSeis2DMatToSegyWithRef(seis2dmat, seisfile,
                                refsegy,
                                datacol=3, verbose=True,
                                qpgsdlg=None):
    """
    Write seismic data from a 2D matrix to segy file, with survey information from a reference segy file

    Args:
        seis2dmat:  seismic data in a 2D matrix containing at least one columns [Value, ...]
                    Default setting is for a 4-column 2D matrix [IL, XL, Z, Value]
                    Adjust targetcol correspondingly.
        seisfile:   name of the output segy file
        refsefy:    reference segy file for copying the header information
        datacol:    index of the data column for output (indexing from 0)
                    Default is 3 (the fourth column)
        verbose:    flag for message display. Default is True
        qpgsdlg:    QProgressDialog for displaying progress. Default is None

    Returns:
        N/A
    """

    # Check if file exist
    if not os.path.isfile(refsegy):
        vis_msg.print('ERROR in writeSeis2DMatToSegyWithRef: ' + refsegy + ' not found', type='error')
        sys.exit()

    # Check size of input 2D matrix
    if np.ndim(seis2dmat) != 2:
        vis_msg.print('ERROR in writeSeis2DMatToSegyWithRef: 2D seismic matrix expected', type='error')
        sys.exit()
    if datacol < 0 or len(seis2dmat[0, :]) <= datacol:
        vis_msg.print('ERROR in writeSeis2DMatToSegyWithRef: No data column found in 2D seismic matrix', type='error')
        sys.exit()

    # Extract the output information: the fourth column
    seis2dmat = seis2dmat[:, datacol]

    with open(refsegy, 'rb') as refid, \
            open(seisfile, 'wb') as sid:
        seisdata = syreader.create_reader(refid)

        # Get the sample number per trace from ref segy
        znum = seisdata.trace_header(0).num_samples

        samplenum = len(seis2dmat)
        tracenum = int(samplenum / znum)

        seis2dmat = seis2dmat.transpose()
        seis2dmat = np.reshape(seis2dmat, [tracenum, znum])
        seis2dmat = seis2dmat.transpose()

        # Write 3200-byte textural header
        sywriter.write_textual_reel_header(sid,
                                           seisdata.textual_reel_header,
                                           encoding=seisdata.encoding)
        # Write 400-byte binary header
        sywriter.write_binary_reel_header(sid,
                                          seisdata.binary_reel_header,
                                          endian=seisdata.endian)
        # Write extended textural header
        sywriter.write_extended_textual_headers(sid,
                                                seisdata.extended_textual_header,
                                                encoding=seisdata.encoding)

        # Prepare trace header format
        traceheaderpacker = sypacker.HeaderPacker(TraceHeaderRev1,
                                                  endian=seisdata.endian)

        if qpgsdlg is not None:
            qpgsdlg.setMaximum(tracenum)
        for itrace in seisdata.trace_indexes():
            #
            if qpgsdlg is not None:
                QtCore.QCoreApplication.instance().processEvents()
                qpgsdlg.setValue(itrace)
            #
            if verbose:
                sys.stdout.write(
                    '\r>>> Write %d traces into SEG-Y: proceeding %.1f%% ' % (
                    tracenum, float(itrace + 1) / float(tracenum) * 100))
                sys.stdout.flush()
                # print('writing ' + str(itrace+1) + ' of ' + str(tracenum) + ' traces')
            # Write 240-byte trace header one-by-one
            sywriter.write_trace_header(sid,
                                        seisdata.trace_header(itrace),
                                        trace_header_packer=traceheaderpacker)
            # Write trace data after each trace header
            sywriter.write_trace_samples(sid,
                                         seis2dmat[:, itrace],
                                         seisdata.data_sample_format,
                                         endian=seisdata.endian)

            if itrace >= tracenum:
                vis_msg.print('WARNING in writeSeis2DMatToSegyWithRef: Reference segy file ' + refsegy +
                              ' contains more traces than 2D seismic matrix', type='warning')
                break
        if verbose:
            print('Done --> output file saved as ' + seisfile)
        #
        if qpgsdlg is not None:
            qpgsdlg.setValue(tracenum)

    return True


def writeSeis3DMatToSegyWithRef(seis3dmat, seisfile,
                                refsegy,
                                verbose=True,
                                qpgsdlg=None):
    """
    Write seismic data from a 3D matrix to segy file, with survey information from a reference segy file

    Args:
        seis3dmat:  seismic data in a 3D matrix [Z/Crossline/Inline]
        seisfile:   name of the output segy file
        refsegy:    reference segy file for copying the header information
        verbose:    flag for message display. Default is True
        qpgsdlg:    QProgressDialog for displaying progress. Default is None

    Returns:
        N/A
    """

    # Check if file exist
    if os.path.isfile(refsegy) == False:
        vis_msg.print('ERROR in writeSeis3DMatToSegyWithRef: ' + refsegy + ' not found', type='error')
        sys.exit()

    # Check the dimension of input
    if np.ndim(seis3dmat) != 3:
        vis_msg.print('ERROR in writeSeis3DMatToSegyWithRef: 3D seismic matrix expected', type='error')
        sys.exit()

    with open(refsegy, 'rb') as refid, \
        open(seisfile, 'wb') as sid:
        seisdata = syreader.create_reader(refid)

        znum, xlnum, inlnum = np.shape(seis3dmat)
        znum = np.round(znum).astype(np.int32)
        xlnum = np.round(xlnum).astype(np.int32)
        inlnum = np.round(inlnum).astype(np.int32)
        tracenum = (inlnum * xlnum).astype(np.int64)

        seis3dmat = seis3dmat.transpose()
        seis3dmat = np.reshape(seis3dmat, [inlnum*xlnum, znum])
        seis3dmat = seis3dmat.transpose()

        # Write 3200-byte textural header
        sywriter.write_textual_reel_header(sid,
                                           seisdata.textual_reel_header,
                                           encoding=seisdata.encoding)
        # Write 400-byte binary header
        sywriter.write_binary_reel_header(sid,
                                          seisdata.binary_reel_header,
                                          endian=seisdata.endian)
        # Write extended textural header
        sywriter.write_extended_textual_headers(sid,
                                                seisdata.extended_textual_header,
                                                encoding=seisdata.encoding)

        # Prepare trace header format
        traceheaderpacker = sypacker.HeaderPacker(TraceHeaderRev1,
                                                  endian=seisdata.endian)

        if qpgsdlg is not None:
            qpgsdlg.setMaximum(tracenum)
        for itrace in seisdata.trace_indexes():
            #
            if qpgsdlg is not None:
                QtCore.QCoreApplication.instance().processEvents()
                qpgsdlg.setValue(itrace)
            #
            if verbose:
                sys.stdout.write(
                    '\r>>> Write %d traces into SEG-Y: proceeding %.1f%%' % (tracenum, float(itrace + 1) / float(tracenum) * 100))
                sys.stdout.flush()
                # print('writing ' + str(itrace+1) + ' of ' + str(tracenum) + ' traces')
            # Write 240-byte trace header one-by-one
            sywriter.write_trace_header(sid,
                                        seisdata.trace_header(itrace),
                                        trace_header_packer=traceheaderpacker)
            # Write trace data after each trace header
            sywriter.write_trace_samples(sid,
                                         seis3dmat[:, itrace],
                                         seisdata.data_sample_format,
                                         endian=seisdata.endian)

            if itrace >= tracenum:
                vis_msg.print('WARNING in writeSeis3DMatToSegyWithRef: Reference segy file ' + refsegy +
                              ' contains more traces than input 3D seismic matrix', type='warning')
                break
        if verbose:
            print('Done --> output file saved as ' + seisfile)
        #
        if qpgsdlg is not None:
            qpgsdlg.setValue(tracenum)

    return True


def writeSeis2DMatToSegyNoRef(seis2dmat, seisfile,
                              xcoorstart=0.0, ycoorstart=0.0,
                              inlbinsize=1.0, xlbinsize=1.0, orientation=90,
                              inlcol=0, xlcol=1, zcol=2,
                              datacol=3, verbose=True,
                              qpgsdlg=None):
    """
    Write seismic data from a 3D matrix to segy file, with user-defined survey information

    Args:
        seis2dmat:      seismic data in a 2D matrix containing at least one columns [Value, ...]
                        Default setting is for a 4-column 2D matrix [IL, XL, Z, Value]
                        Adjust targetcol correspondingly.
        seisfile:       name of the output segy file
        xcoorstart:     x coordinate of first trace. Default is 0.0
        ycoorstart:     y coordinate of first trace. Default is 0.0
        inlbinsize:     inline bin size. Default is 1.0
        xlbinsize:      crossline bin size. Default is 1.0
        orientation:    orienation of the first crossline from N direction in degree. Default is 90
        inlcol:         index of inline column. Default is the first column (0)
        xlcol:          index of crossline column. Default is the second column (1)
        zcol:           index of z column. Default is the third column (2)
        datacol:        index of the data column for output (indexing from 0)
                        Default is 3 (the fourth column)
        verbose:        flag for message display. Default is True
        qpgsdlg:        QProgressDialog for displaying progress. Default is None

    Returns:
        N/A

    Note:
        Negative z is used in the vertical direction
        ASCII is used for textual header
        IBM floating is used for data format
    """

    # Check size of input 2D matrix
    if np.ndim(seis2dmat) != 2:
        vis_msg.print('ERROR in writeSeis2DMatToSegyNoRef: 2D seismic matrix expected', type='error')
        sys.exit()
    if datacol < 0 or len(seis2dmat[0, :]) <= datacol:
        vis_msg.print('ERROR in writeSeis2DMatToSegyNoRef: No data column found in 2D seismic matrix',
                      type='error')
        sys.exit()

    # Extract the output information: the fourth column
    seis3dmat = seis_ays.convertSeis2DMatTo3DMat(seis2dmat, datacol=datacol)
    seisinfo = seis_ays.getSeisInfoFrom2DMat(seis2dmat, inlcol=inlcol, xlcol=xlcol, zcol=zcol)
    inlstart = seisinfo['ILStart']
    inlstep = seisinfo['ILStep']
    xlstart = seisinfo['XLStart']
    xlstep = seisinfo['XLStep']
    zstart = seisinfo['ZStart']
    zstep = seisinfo['ZStep']

    with open(seisfile, 'wb') as sid:

        znum, xlnum, inlnum = np.shape(seis3dmat)
        znum = np.round(znum).astype(np.int32)
        xlnum = np.round(xlnum).astype(np.int32)
        inlnum = np.round(inlnum).astype(np.int32)
        tracenum = (inlnum * xlnum).astype(np.int64)

        seis3dmat = seis3dmat.transpose()
        seis3dmat = np.reshape(seis3dmat, [inlnum * xlnum, znum])
        seis3dmat = seis3dmat.transpose()

        # Write 3200-byte textural header
        textualheader = createSegyTextualHeader()
        sywriter.write_textual_reel_header(sid,
                                           textualheader,
                                           encoding='ascii')
        # Write 400-byte binary header
        binaryheader = createSegyBinaryHeader(znum, zstep)
        sywriter.write_binary_reel_header(sid,
                                          binaryheader)
        # Write extended textural header (No)

        # Prepare trace header format
        traceheaderpacker = sypacker.HeaderPacker(TraceHeaderRev1)

        if qpgsdlg is not None:
            qpgsdlg.setMaximum(tracenum)
        for itrace in range(tracenum):
            #
            if qpgsdlg is not None:
                QtCore.QCoreApplication.instance().processEvents()
                qpgsdlg.setValue(itrace)
            #
            if verbose:
                sys.stdout.write(
                    '\r>>> Write %d traces into SEG-Y: proceeding %.1f%%' % (
                    tracenum, float(itrace + 1) / float(tracenum) * 100))
                sys.stdout.flush()
                # print('writing ' + str(itrace+1) + ' of ' + str(tracenum) + ' traces')
            # Write 240-byte trace header one-by-one
            xcoor = xcoorstart - int(itrace/xlnum) * inlbinsize * math.cos(orientation*math.pi/180)
            ycoor = ycoorstart + int(itrace/xlnum) * inlbinsize * math.sin(orientation*math.pi/180)
            traceheader = createSegyTraceHeader(inlstart+int(itrace/xlnum)*inlstep, xlstart+int(itrace%xlnum)*xlstep,
                                                xcoor+int(itrace%xlnum)*xlbinsize*math.sin(orientation*math.pi/180),
                                                ycoor+int(itrace%xlnum)*xlbinsize*math.cos(orientation*math.pi/180),
                                                znum, zstep, zstart)
            sywriter.write_trace_header(sid,
                                        traceheader,
                                        trace_header_packer=traceheaderpacker)
            # Write trace data after each trace header
            sywriter.write_trace_samples(sid,
                                         seis3dmat[:, itrace],
                                         'ibm')
        if verbose:
            print('Done --> output file saved as ' + seisfile)
        #
        if qpgsdlg is not None:
            qpgsdlg.setValue(tracenum)

    return True


def writeSeis3DMatToSegyNoRef(seis3dmat, seisfile,
                              xcoorstart=0.0, ycoorstart=0.0,
                              inlbinsize=1.0, xlbinsize=1.0, orientation=90,
                              inlstart=1, inlstep=1, xlstart=1, xlstep=1,
                              zstart=0, zstep=-1,
                              verbose=True,
                              qpgsdlg=None):
    """
    Write seismic data from a 3D matrix to segy file, with user-defined survey information

    Args:
        seis3dmat:      seismic data in a 3D matrix [Z/Crossline/Inline]
        seisfile:       name of the output segy file
        xcoorstart:     x coordinate of first trace. Default is 0.0
        ycoorstart:     y coordinate of first trace. Default is 0.0
        inlbinsize:     inline bin size. Default is 1.0
        xlbinsize:      crossline bin size. Default is 1.0
        orientation:    orienation of the first crossline from N direction in degree. Default is 90
        inlstart:       inline number of first trace. Default is 1
        inlstep:        inline step. Default is 1
        xlstart:        crossline number of first trace. Default is 1
        xlstep:         crossline step. Default is 1
        zstart:         z value of first sample. Default is 0
        zstep:          z step. Default is -1
        verbose:        flag for message display. Default is True
        qpgsdlg:        QProgressDialog for displaying progress. Default is None

    Returns:
        N/A

    Note:
        Negative z is used in the vertical direction
        ASCII is used for textual header
        IBM floating is used for data format
    """

    # Check the dimension of input
    if np.ndim(seis3dmat) != 3:
        vis_msg.print('ERROR in writeSeis3DMatToSegyNoRef: 3D seismic matrix expected', type='error')
        sys.exit()

    with open(seisfile, 'wb') as sid:
        znum, xlnum, inlnum = np.shape(seis3dmat)
        znum = np.round(znum).astype(np.int32)
        xlnum = np.round(xlnum).astype(np.int32)
        inlnum = np.round(inlnum).astype(np.int32)
        tracenum = (inlnum * xlnum).astype(np.int64)

        seis3dmat = seis3dmat.transpose()
        seis3dmat = np.reshape(seis3dmat, [inlnum * xlnum, znum])
        seis3dmat = seis3dmat.transpose()

        # Write 3200-byte textural header
        textualheader = createSegyTextualHeader()
        sywriter.write_textual_reel_header(sid,
                                           textualheader,
                                           encoding='ascii')
        # Write 400-byte binary header
        binaryheader = createSegyBinaryHeader(znum, zstep)
        sywriter.write_binary_reel_header(sid,
                                          binaryheader)
        # Write extended textural header (No)

        # Prepare trace header format
        traceheaderpacker = sypacker.HeaderPacker(TraceHeaderRev1)


        if qpgsdlg is not None:
            qpgsdlg.setMaximum(tracenum)
        for itrace in range(tracenum):
            #
            if qpgsdlg is not None:
                QtCore.QCoreApplication.instance().processEvents()
                qpgsdlg.setValue(itrace)
            #
            if verbose:
                sys.stdout.write(
                    '\r>>> Write %d traces into SEG-Y: proceeding %.1f%%' % (
                    tracenum, float(itrace + 1) / float(tracenum) * 100))
                sys.stdout.flush()
                # print('writing ' + str(itrace+1) + ' of ' + str(tracenum) + ' traces')
            # Write 240-byte trace header one-by-one
            xcoor = xcoorstart - int(itrace/xlnum) * inlbinsize * math.cos(orientation*math.pi/180)
            ycoor = ycoorstart + int(itrace/xlnum) * inlbinsize * math.sin(orientation*math.pi/180)
            traceheader = createSegyTraceHeader(inlstart+int(itrace/xlnum)*inlstep, xlstart+int(itrace%xlnum)*xlstep,
                                                xcoor+int(itrace%xlnum)*xlbinsize*math.sin(orientation*math.pi/180),
                                                ycoor+int(itrace%xlnum)*xlbinsize*math.cos(orientation*math.pi/180),
                                                znum, zstep, zstart)
            sywriter.write_trace_header(sid,
                                        traceheader,
                                        trace_header_packer=traceheaderpacker)
            # Write trace data after each trace header
            sywriter.write_trace_samples(sid,
                                         seis3dmat[:, itrace],
                                         'ibm')
        if verbose:
            print('Done --> output file saved as ' + seisfile)
        #
        if qpgsdlg is not None:
            qpgsdlg.setValue(tracenum)

    return True


def createSegyTextualHeader():
    """
    Create 3200-byte Seg-Y textual header

    Args:
        None

    Returns:
        header: A tuple of forty Unicode strings containing the transcoded header data.
    """

    header = ()
    header = header + ('C01 Created By: Cognitive Geo                                                   ',)
    header = header + ('C02 Author: GeoPy Team (geopy.info@gmail.com)                                   ',)
    header = header + ('C03                                                                             ',)
    header = header + ('C04                                                                             ',)
    header = header + ('C05                                                                             ',)
    header = header + ('C06                                                                             ',)
    header = header + ('C07                                                                             ',)
    header = header + ('C08                                                                             ',)
    header = header + ('C09                                                                             ',)
    header = header + ('C10                                                                             ',)
    header = header + ('C11 Key Parameters:                                                             ',)
    header = header + ('C12 a. 4-byte IBM Floating-point Data Format                                    ',)
    header = header + ('C13 b. SEG-Y Format Revision 1.0                                                ',)
    header = header + ('C14 c. Fixed Length for All Traces                                              ',)
    header = header + ('C15 d. No 3200-byte Extended Textual Header                                     ',)
    header = header + ('C16                                                                             ',)
    header = header + ('C17                                                                             ',)
    header = header + ('C18                                                                             ',)
    header = header + ('C19                                                                             ',)
    header = header + ('C20                                                                             ',)
    header = header + ('C21 Key Header Information:                                                     ',)
    header = header + ('C22 a. Inline No. at Bytes 9 & 189                                              ',)
    header = header + ('C23 b. Crossline No. at Bytes 21 & 193                                          ',)
    header = header + ('C24 c. X Coordinate at Bytee 73 & 181                                           ',)
    header = header + ('C25 d. Y Coordinate at Bytes 77 & 185                                           ',)
    header = header + ('C26 e. Number of Samples per Trace at Byte 115                                  ',)
    header = header + ('C27 f. Sampling Interval in Microseconds at Byte 117                            ',)
    header = header + ('C28                                                                             ',)
    header = header + ('C29                                                                             ',)
    header = header + ('C30                                                                             ',)
    header = header + ('C31                                                                             ',)
    header = header + ('C32                                                                             ',)
    header = header + ('C33                                                                             ',)
    header = header + ('C34                                                                             ',)
    header = header + ('C35                                                                             ',)
    header = header + ('C36                                                                             ',)
    header = header + ('C37                                                                             ',)
    header = header + ('C38                                                                             ',)
    header = header + ('C39                                                                             ',)
    header = header + ('C40 End of Textual Header                                                       ',)

    return header


def createSegyBinaryHeader(znum, zstep):
    """
    Create 400-byte Segy-Y binary header

    Args:
        znum:   number of samples per trace. (offset=3221)
        zstep:  sample interval in millisecond. (offset=3217)

    Returns:
        header: Binary header information stored as class BinaryReelHeader

    Note:
        Negative z is used in the vertical direction
        IBM floating is used as data format [1]
    """

    if znum <= 0:
        vis_msg.print('ERROR in createSegyBinaryHeader: Positive num of samples per trace expected',
                      type='error')
        sys.exit()
    if zstep >= 0:
        vis_msg.print('ERROR in createSegyBinaryHeader: Negative sample interval in milliseconds (ms) expected',
                      type='error')
        sys.exit()

    header = BinaryReelHeader()
    header.num_samples = znum
    header.sample_interval = - zstep * 1000

    header.data_sample_format = 1
    header.format_revision_num = 256
    header.fixed_length_trace_flag = 1
    header.num_extended_textual_headers = 0

    return header


def createSegyTraceHeader(inlidx, xlidx, xcoor, ycoor,
                          znum, zstep, zlag):
    """
    Create 240-byte Seg-Y trace header

    Args:
        inlidx: inline number. (offset = 9 and 189)
        xlidx:  crossline number. (offset = 21 and 193)
        xcoor:  x-coordinate (offset = 73 and 181)
        ycoor:  y-coordinate (offset = 77 and 185)
        znum:   number of samples per trace. (offset = 115)
        zstep:  sample interval in millisecond. (offset = 117)
        zlag:   vertical z lag-time in millisecond. (offset = 105 and 109)

    Return:
        header: Trace header information stored as class TraceHeaderRev1

    Note:
        Negative z is used in the vertical direction
    """

    if inlidx < 0:
        vis_msg.print('ERROR in createSegyTraceHeader: Non-negative inline number expected', type='error')
        sys.exit()
    if xlidx < 0:
        vis_msg.print('ERROR in createSegyTraceHeader: Non-negative crossline number expected', type='error')
        sys.exit()
    if znum <= 0:
        vis_msg.print('ERROR in createSegyTraceHeader: Positive num of samples per trace expected', type='error')
        sys.exit()
    if zstep >= 0:
        vis_msg.print('ERROR in createSegyTraceHeader: Negative sample interval in milliseconds (ms) expected',
                      type='error')
        sys.exit()
    if zlag > 0:
        vis_msg.print('ERROR in createSegyTraceHeader: Non-positive vertical lag-time in millisecond (ms) expected',
                      type='error')
        sys.exit()

    header = TraceHeaderRev1()
    # 9 - 4 byte
    header.field_record_num = inlidx
    # 21 - 4 byte
    header.ensemble_num = xlidx
    # 73 - 4 byte
    header.source_x = xcoor
    # 77 - 4 byte
    header.source_y = ycoor
    # 105 - 2 byte
    header.lag_time_a = zlag
    # 109 - 2 byte
    header.delay_recording_time = - zlag
    # 115 - 2 byte
    header.num_samples = znum
    # 117 - 2 byte
    header.sample_interval = - zstep * 1000
    # 181 - 4 byte
    header.cdp_x = xcoor
    # 185 - 4 byte
    header.cdp_y = ycoor
    # 189 - 4 byte
    header.inline_number = inlidx
    # 193 - 4 byte
    header.crossline_number = xlidx

    return header


def defSegyTraceHeaderFormat(x_byte=181, y_byte=185,
                            inl_byte=189, xl_byte=193):
    """
    Define trace hearder format with the given keywords

    Args:
        x_byte:     byte of x coordinate. Default is 181
        y_byte:     byte of y coordinate. Default is 185
        inl_byte:   byte of inline number. Default is 189
        xl_byte:    byte of crossline number. Default is 193

    Return:
        The defined trace header format
    """

    class SegyTraceHeaderFormat(TraceHeaderRev1):
        cdp_x = field(
            Int32, offset=x_byte, default=0, documentation=
            "X coordinate of ensemble (CDP) position of this trace. "
            "The coordinate reference system should be identified through an extended header Location Data stanza. "
            "The xy_scalar field applies to this value."
        )

        cdp_y = field(
            Int32, offset=y_byte, default=0, documentation=
            "Y coordinate of ensemble (CDP) position of this trace. "
            "The coordinate reference system should be identified through an extended header Location Data stanza. "
            "The xy_scalar field applies to this value."
        )

        inline_number = field(
            Int32, offset=inl_byte, default=0, documentation=
            "In-line number for 3-D poststack data. If one in-line per SEG Y file is "
            "being recorded, this value should be the same for all traces in the file and the same value will be recorded "
            "in bytes 3205-3208 of the Binary File Header."  # TODO: replace bytes with field name
        )

        crossline_number = field(
            Int32, offset=xl_byte, default=0, documentation=
            "Cross-line number for 3-D poststack data. This will typically be the same "
            "value as the ensemble (CDP) number in Trace Header ensemble_num field, but this does not have to be the case."
        )
    return SegyTraceHeaderFormat


def readSegyTextualHeader(segyfile, encoding=None):
    """
    Read textual header from a segy file

    Args:
        segyfile:   segy file
        encoding:   encoding

    Return:
        A tuple of forty Unicode strings containing the transcoded header data
    """

    # Check if file exist
    if os.path.isfile(segyfile) == False:
        vis_msg.print('ERROR in readSegyTextualHeader: ' + segyfile + ' not found', type='error')
        sys.exit()
    with open(segyfile, 'rb') as sid:
        if encoding is None:
            encoding = syreader.guess_textual_header_encoding(sid)
        header = syreader.read_textual_reel_header(sid, encoding=encoding)
    return header


def readSegyBinaryHeader(segyfile, endian='>'):
    """
    Read binary header from a segy file

    Args:
        segyfile:   segy file
        endian:     endian. Default is '>'

    Return:
        A binary header object (defined in segpy)
    """

    # Check if file exist
    if os.path.isfile(segyfile) == False:
        vis_msg.print('ERROR in readSegyBinaryHeader: ' + segyfile + ' not found', type='error')
        sys.exit()
    with open(segyfile, 'rb') as sid:
        header = syreader.read_binary_reel_header(sid, endian=endian)
    return header


def readSegyTraceHeader(segyfile, traceidx=0, traceheaderformat=TraceHeaderRev1, endian='>'):
    """
    Read trace header at a given trace from a segy file

    Args:
        segyfile:           segy file
        traceidx:           index of the target trace. Default is 0
        traceheaderformat:  trace header format. Default is TraceHeaderRev1
        endian:             endian. Default is '>'

    Return:
        A trace header object (defined in segpy)
    """

    # Check if file exist
    if os.path.isfile(segyfile) == False:
        vis_msg.print('ERROR in readSegyTraceHeader: ' + segyfile + ' not found', type='error')
        sys.exit()
    #
    traceheaderpacker = sypacker.HeaderPacker(traceheaderformat, endian)
    pos = np.int64(3200 + 400)
    dataformat = np.array([4, 4, 2, 4, 4, 0, 0, 1])
    with open(segyfile, 'rb') as sid:
        databyte = syreader.read_binary_reel_header(sid, endian=endian)
        databyte = dataformat[databyte.data_sample_format-1]
        for i in range(traceidx):
            header = syreader.read_trace_header(sid, trace_header_packer=traceheaderpacker, pos=pos)
            pos = pos + 240 + header.num_samples * databyte
        header = syreader.read_trace_header(sid, trace_header_packer=traceheaderpacker, pos=pos)
    return header


class inputoutput:
    # pack all functions as a class
    #
    readSeisInfoFromSegy = readSeisInfoFromSegy
    readSeisDataFromSegy = readSeisDataFromSegy
    readSeis2DMatFromSegyWithInfo = readSeis2DMatFromSegyWithInfo
    readSeis3DMatFromSegyWithInfo = readSeis3DMatFromSegyWithInfo
    readSeis2DMatFromSegyNoInfo = readSeis2DMatFromSegyNoInfo
    readSeis3DMatFromSegyNoInfo = readSeis3DMatFromSegyNoInfo
    #
    writeSeis2DMatToSegyWithRef = writeSeis2DMatToSegyWithRef
    writeSeis3DMatToSegyWithRef = writeSeis3DMatToSegyWithRef
    writeSeis2DMatToSegyNoRef = writeSeis2DMatToSegyNoRef
    writeSeis3DMatToSegyNoRef = writeSeis3DMatToSegyNoRef
    #
    createSegyTextualHeader = createSegyTextualHeader
    createSegyBinaryHeader = createSegyBinaryHeader
    createSegyTraceHeader = createSegyTraceHeader
    #
    defSegyTraceHeaderFormat = defSegyTraceHeaderFormat
    #
    readSegyTextualHeader = readSegyTextualHeader
    readSegyBinaryHeader = readSegyBinaryHeader
    readSegyTraceHeader = readSegyTraceHeader


class Int16(int):

    MINIMUM = -32768
    MAXIMUM = 32767
    SIZE = 2
    SEG_Y_TYPE = 'int16'

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls, *args, **kwargs)
        if not (Int16.MINIMUM <= instance <= Int16.MAXIMUM):
            raise ValueError("{} value {!r} outside range {} to {}".format(cls.__name__, instance,
                                                                        cls.MINIMUM, cls.MAXIMUM))
        return instance


class Int32(int):

    MINIMUM = -2147483648
    MAXIMUM = 2147483647
    SIZE = 4
    SEG_Y_TYPE = 'int32'

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls, *args, **kwargs)
        if not (Int32.MINIMUM <= instance <= Int32.MAXIMUM):
            raise ValueError("{} value {!r} outside range {} to {}".format(cls.__name__, instance,
                                                                        cls.MINIMUM, cls.MAXIMUM))
        return instance




#############################################################################################
#                                                                                           #
# Author:   GeoPy Team                                                                      #
# Email:    geopy.info@gmail.com                                                            #
# Date:     March 2018                                                                      #
#                                                                                           #
#############################################################################################

# basic functions for processing video

import sys
import os
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation
#
sys.path.append(os.path.dirname(__file__)[:-4][:-4][:-13])
from cognitivegeo.src.vis.colormap import colormap as vis_cmap


__all__ = ['video']


def animateVideo(video, video_height, video_width, video_depth,
                 frame_list=[],
                 cmapname='Red-White-Blue', flipcmap=False, fontsize=12,
                 savepath='./', savename='video.mp4',
                 verbose=True):
    """
    Generate animation of a single video (by matplotlib.animation)

    Args:
        video:          2D matrix of videos [1, npixel]
        video_height:   height of video
        video_width:    width of video
        video_depth:    depth of video
        frame_list:     list of frames. Default is [] for all frames
        cmapname:       name of colormap as listed below:
                            'seismic', 'phase', 'frequency',
                            'red_white_blue',
                            'black_white_red', 'black_white_green', 'black_white_blue',
                            'white_gray_black',
                            'white_red_black', 'white_green_black', 'white_blue_black',
                            'black_red', 'black_green', 'black_blue'
                        Default is 'red_white_blue'
        flipcmap:       Flip colormap. Default is False
        fontsize:       font size
        savepath:       path of saving the animation. Default is current directory
        savename:       name of saving the animation. Default is tfgan_er.mp4
        verbose:        display the animation process. Default is True

    Returns:
        None
    """

    if np.ndim(video) != 2:
        print('ERROR in animateVideo: 2D video matrix expected')
        sys.exit()
    if video_height <= 1 or video_width <= 1 or video_depth <= 1:
        print('ERROR in animateVideo: Video height/width/depth be > 1')
        sys.exit()

    if os.path.exists(savepath) is False:
        print('ERROR in animateVideo: No directory found for saving animation file')
        sys.exit()
    if os.path.exists(os.path.join(savepath, savename)) is True:
        print('WARNING in animateVideo: Overwrite pre-existing animation file')

    nvideo, npixel = np.shape(video)

    if nvideo > 1:
        print('WARNING in animateVideo: Only first video selected')
        video = video[0, :]

    if npixel != video_height * video_width * video_depth:
        print('ERROR in animateVideo: Video height/width/depth not match')
        sys.exit()

    if len(frame_list) < 1 or np.min(frame_list) < 0 or np.max(frame_list) > video_depth-1:
        print('WARNING in animateVideo: All frames selected')
        frame_list = np.linspace(0, video_depth - 1, video_depth, dtype=np.int)

    frame_list = frame_list.astype(int)

    video = np.reshape(video, [video_height, video_width, video_depth])

    fig = plt.figure(facecolor='white')

    def animate(i):
        if verbose and (i + 1) % 10 == 0:
            print('Animate ' + str(i + 1) + ' of ' + str(len(frame_list)) + ' frames')
        fig.clear()
        fig.suptitle('Frame No. = ' + str(frame_list[i]))
        plt.rc('font', size=fontsize)
        ims = []
        im = plt.imshow(video[:, :, frame_list[i]],
                        interpolation='bicubic',
                        cmap=vis_cmap.makeColorMap(cmapname, flipcmap))
        plt.axis('off')
        ims.append(im)
        return ims

    anim = animation.FuncAnimation(fig, animate, frames=len(frame_list), interval=500,
                                   blit=True, init_func=None)

    writer = animation.FFMpegWriter()
    anim.save(os.path.join(savepath, savename), writer=writer, dpi=300)

    return True


def animateVideoGallery(video, video_height, video_width, video_depth,
                        frame_list=[], ncol = 5,
                        cmapname='red_white_blue', flipcmap=False,
                        spacing=False, maxfigsize=10, fontsize=12,
                        savepath='./', savename='video.mp4',
                        verbose=True):
    """
    Generate animation of multiple videos (by matplotlib.animation)

    Args:
        video:          2D matrix of videos [nvideo, npixel]
        video_height:   height of video
        video_width:    width of video
        video_depth:    depth of video
        frame_list:     list of frames. Default is [] for all frames
        ncol:           number of videos per row
        cmapname:       name of colormap as listed below:
                            'seismic', 'phase', 'frequency',
                            'red_white_blue',
                            'black_white_red', 'black_white_green', 'black_white_blue',
                            'white_gray_black',
                            'white_red_black', 'white_green_black', 'white_blue_black',
                            'black_red', 'black_green', 'black_blue'
                        Default is 'red_white_blue'
        flipcmap:       Flip colormap. Default is False
        spacing:        spacing or not between subplots. Default is False
        maxfigsize:     maximum figure size. Default is 10
        fontsize:       font size. Default is 12
        savepath:       path of saving the animation. Default is current directory
        savename:       name of saving the animation. Default is tfgan_er.mp4
        verbose:        display the animation process. Default is True

    Returns:
        None
    """

    if np.ndim(video) != 2:
        print('ERROR in animateVideoGallery: 2D video matrix expected')
        sys.exit()
    if video_height <= 1 or video_width <= 1 or video_depth <= 1:
        print('ERROR in animateVideoGallery: Video height/width/depth be > 1')
        sys.exit()

    if os.path.exists(savepath) is False:
        print('ERROR in animateVideoGallery: No directory found for saving animation file')
        sys.exit()
    if os.path.exists(os.path.join(savepath, savename)) is True:
        print('WARNING in animateVideoGallery: Overwrite pre-existing animation file')

    nvideo, npixel = np.shape(video)

    if ncol < 1:
        print('ERROR in animateVideoGallery: Column number be >= 1')
        sys.exit()

    nrow = int(nvideo / ncol)

    if npixel != video_height * video_width * video_depth:
        print('ERROR in animateVideoGallery: Video height/width/depth not match')
        sys.exit()

    if len(frame_list) < 1 or np.min(frame_list) < 0 or np.max(frame_list) > video_depth-1:
        print('WARNING in animateVideoGallery: All frames selected')
        frame_list = np.linspace(0, video_depth - 1, video_depth, dtype=np.int)

    frame_list = frame_list.astype(int)

    video = np.reshape(video,[nvideo, video_height, video_width, video_depth])

    figwidth = float(ncol * video_width)
    figheight = float(nrow * video_height)
    if figwidth > figheight:
        figheight = figheight / figwidth * maxfigsize
        figwidth = maxfigsize
    else:
        figwidth = figwidth / figheight * maxfigsize
        figheight = maxfigsize

    fig = plt.figure(facecolor='white', figsize=(figwidth, figheight))

    # initialization function: plot the background of each frame
    def init():
        ims = []
        for k in range(nrow * ncol):
            plt.subplot(nrow, ncol, k + 1)
            im = plt.imshow(np.zeros([video_height, video_width]),
                            interpolation='bicubic',
                            cmap=vis_cmap.makeColorMap(cmapname, flipcmap))
            plt.axis('off')
            ims.append(im)
        return ims

    def animate(i):
        if verbose and (i + 1) % 10 == 0:
            print('Animate ' + str(i + 1) + ' of ' + str(len(frame_list)) + ' frames')
        fig.clear()
        fig.suptitle('Frame No. = ' + str(frame_list[i]))
        plt.rc('font', size=fontsize)
        ims = []
        for k in range(nrow*ncol):
            plt.subplot(nrow, ncol, k+1)
            im = plt.imshow(video[k, :, :, frame_list[i]],
                            interpolation='bicubic',
                            cmap=vis_cmap.makeColorMap(cmapname, flipcmap))
            plt.axis('off')
            ims.append(im)
        if spacing is False:
            plt.subplots_adjust(wspace=0, hspace=0)
        return ims

    anim = animation.FuncAnimation(fig, animate, frames=len(frame_list), interval=500,
                                   blit=True, init_func=None)

    writer = animation.FFMpegWriter()
    anim.save(os.path.join(savepath, savename), writer=writer, dpi=300)

    return True


class video:
    # Pack all functions as a class
    #
    animateVideo = animateVideo
    animateVideoGallery = animateVideoGallery


#############################################################################################
#                                                                                           #
# Author:   GeoPy Team                                                                      #
# Email:    geopy.info@gmail.com                                                            #
# Date:     March 2018                                                                      #
#                                                                                           #
#############################################################################################

# Create a window for editing seismic/pointset properties


from PyQt5 import QtCore, QtGui, QtWidgets
import os, sys
#
sys.path.append(os.path.dirname(__file__)[:-4][:-4][:-13])
from cognitivegeo.src.pointset.analysis import analysis as point_ays
from cognitivegeo.src.vis.messager import messager as vis_msg
from cognitivegeo.src.gui.rtrvseisprop import rtrvseisprop as gui_rtrvseisprop
from cognitivegeo.src.gui.calculator import calculator as gui_calculator

QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)
QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)


class editseispointset(object):

    seispointdata = {}
    rootpath = ''
    #
    iconpath = os.path.dirname(__file__)
    dialog = None


    def setupGUI(self, EditSeisPointSet):
        EditSeisPointSet.setObjectName("EditSeisPointSet")
        EditSeisPointSet.setFixedSize(300, 420)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(os.path.join(self.iconpath, "icons/seismic.png")),
                       QtGui.QIcon.Normal,
                       QtGui.QIcon.Off)
        EditSeisPointSet.setWindowIcon(icon)
        #
        self.lblattrib = QtWidgets.QLabel(EditSeisPointSet)
        self.lblattrib.setObjectName("lblattrib")
        self.lblattrib.setGeometry(QtCore.QRect(10, 10, 150, 30))
        self.lwgattrib = QtWidgets.QListWidget(EditSeisPointSet)
        self.lwgattrib.setObjectName("lwgattrib")
        self.lwgattrib.setGeometry(QtCore.QRect(10, 50, 280, 200))
        self.lwgattrib.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.lblaction = QtWidgets.QLabel(EditSeisPointSet)
        self.lblaction.setObjectName("lblaction")
        self.lblaction.setGeometry(QtCore.QRect(110, 270, 40, 30))
        self.cbbaction = QtWidgets.QComboBox(EditSeisPointSet)
        self.cbbaction.setObjectName("cbbaction")
        self.cbbaction.setGeometry(QtCore.QRect(160, 270, 130, 30))
        self.ldtrename = QtWidgets.QLineEdit(EditSeisPointSet)
        self.ldtrename.setObjectName("ldtrename")
        self.ldtrename.setGeometry(QtCore.QRect(160, 310, 130, 30))
        self.btnedit = QtWidgets.QPushButton(EditSeisPointSet)
        self.btnedit.setObjectName("btnedit")
        self.btnedit.setGeometry(QtCore.QRect(100, 370, 100, 30))
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(os.path.join(self.iconpath, "icons/ok.png")),
                       QtGui.QIcon.Normal,
                       QtGui.QIcon.Off)
        self.btnedit.setIcon(icon)
        #
        self.msgbox = QtWidgets.QMessageBox(EditSeisPointSet)
        self.msgbox.setObjectName("msgbox")
        _center_x = EditSeisPointSet.geometry().center().x()
        _center_y = EditSeisPointSet.geometry().center().y()
        self.msgbox.setGeometry(QtCore.QRect(_center_x - 150, _center_y - 50, 300, 100))
        #
        self.retranslateGUI(EditSeisPointSet)
        QtCore.QMetaObject.connectSlotsByName(EditSeisPointSet)


    def retranslateGUI(self, EditSeisPointSet):
        self.dialog = EditSeisPointSet
        #
        _translate = QtCore.QCoreApplication.translate
        EditSeisPointSet.setWindowTitle(_translate("EditSeisPointSet", "Edit Seismic/PointSet"))
        self.lblattrib.setText(_translate("EditSeisPointSet", "List of available properties:"))
        self.lblaction.setText(_translate("EditSeisPointSet", "Action:"))
        self.cbbaction.addItems(['Copy', 'Rename', 'Delete', 'Add', 'Math'])
        self.cbbaction.setItemIcon(0, QtGui.QIcon(os.path.join(self.iconpath, "icons/copy.png")))
        self.cbbaction.setItemIcon(1, QtGui.QIcon(os.path.join(self.iconpath, "icons/rename.png")))
        self.cbbaction.setItemIcon(2, QtGui.QIcon(os.path.join(self.iconpath, "icons/delete.png")))
        self.cbbaction.setItemIcon(3, QtGui.QIcon(os.path.join(self.iconpath, "icons/retrieve.png")))
        self.cbbaction.setItemIcon(4, QtGui.QIcon(os.path.join(self.iconpath, "icons/math.png")))
        self.cbbaction.setCurrentIndex(4)
        self.cbbaction.currentIndexChanged.connect(self.changeCbbAction)
        self.ldtrename.setText(_translate("EditSeisPointSet", ""))
        self.ldtrename.setVisible(False)
        self.btnedit.setText(_translate("EditSeisPointSet", "Apply"))
        self.btnedit.clicked.connect(self.clickBtnEditSeisPointSet)
        #
        self.refreshLwgAttrib()


    def changeCbbAction(self):
        if self.cbbaction.currentIndex() == 1:
            self.ldtrename.setVisible(True)
        else:
            self.ldtrename.setText('')
            self.ldtrename.setVisible(False)


    def clickBtnEditSeisPointSet(self):
        self.refreshMsgBox()
        #
        _attriblist = self.lwgattrib.selectedItems()
        if self.cbbaction.currentIndex() != 3 and len(_attriblist) < 1:
            vis_msg.print("ERROR in EditSeisPointSet: No property selected for editing", type='error')
            QtWidgets.QMessageBox.critical(self.msgbox,
                                           'Edit Seismic/PointSet',
                                           'No property selected for editing')
            return
        #
        if self.cbbaction.currentIndex() == 0:
            self.seispointdata[_attriblist[0].text()+'_copy'] = self.seispointdata[_attriblist[0].text()].copy()
        if self.cbbaction.currentIndex() == 1:
            if len(self.ldtrename.text()) < 1:
                vis_msg.print("ERROR in EditSeisPointSet: No name specified for rename", type='error')
                QtWidgets.QMessageBox.critical(self.msgbox,
                                               'Edit Seismic/PointSet',
                                               'No name specified for rename')
                return
            if self.ldtrename.text() in self.seispointdata.keys():
                reply = QtWidgets.QMessageBox.question(self.msgbox, 'Edit Seismic/PointSet',
                                                       self.ldtrename.text() + ' already exists. Overwrite?',
                                                       QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                                                       QtWidgets.QMessageBox.No)

                if reply == QtWidgets.QMessageBox.No:
                    return
            self.seispointdata[self.ldtrename.text()] = self.seispointdata[_attriblist[0].text()]
            self.seispointdata.pop(_attriblist[0].text())
        if self.cbbaction.currentIndex() == 2:
            self.seispointdata.pop(_attriblist[0].text())
        if self.cbbaction.currentIndex() == 3:
            _rtrvprop = QtWidgets.QDialog()
            _gui = gui_rtrvseisprop()
            _gui.pointsetdata = self.seispointdata.copy()
            _gui.rootpath = self.rootpath
            _gui.setupGUI(_rtrvprop)
            _rtrvprop.exec()
            self.seispointdata = _gui.pointsetdata.copy()
            _rtrvprop.show()
        if self.cbbaction.currentIndex() == 4:
            _math = QtWidgets.QDialog()
            _gui = gui_calculator()
            _gui.data = self.seispointdata[_attriblist[0].text()].copy()
            _gui.setupGUI(_math)
            _math.exec()
            self.seispointdata[_attriblist[0].text()] = _gui.data.copy()
            _math.show()
        #
        self.refreshLwgAttrib()
        #
        # QtWidgets.QMessageBox.information(self.msgbox,
        #                                   "Edit Seismic/PointSet",
        #                                   "Seismic/PointSet property edited successfully")
        return


    def refreshLwgAttrib(self):
        self.lwgattrib.clear()
        if self.checkSeisPointSet() is True:
            for i in sorted(self.seispointdata.keys()):
                if i != "Inline" and i != "Crossline" and i != "Z":
                    item = QtWidgets.QListWidgetItem(self.lwgattrib)
                    item.setText(i)
                    self.lwgattrib.addItem(item)
            if "Inline" in self.seispointdata.keys():
                item = QtWidgets.QListWidgetItem(self.lwgattrib)
                item.setText("Inline")
                self.lwgattrib.addItem(item)
            if "Crossline" in self.seispointdata.keys():
                item = QtWidgets.QListWidgetItem(self.lwgattrib)
                item.setText("Crossline")
                self.lwgattrib.addItem(item)
            if "Z" in self.seispointdata.keys():
                item = QtWidgets.QListWidgetItem(self.lwgattrib)
                item.setText("Z")
                self.lwgattrib.addItem(item)


    def refreshMsgBox(self):
        _center_x = self.dialog.geometry().center().x()
        _center_y = self.dialog.geometry().center().y()
        self.msgbox.setGeometry(QtCore.QRect(_center_x - 150, _center_y - 50, 300, 100))


    def checkSeisPointSet(self):
        return point_ays.checkPoint(self.seispointdata)


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    EditSeisPointSet = QtWidgets.QWidget()
    gui = editseispointset()
    gui.setupGUI(EditSeisPointSet)
    EditSeisPointSet.show()
    sys.exit(app.exec_())


#############################################################################################
#                                                                                           #
# Author:   GeoPy Team                                                                      #
# Email:    geopy.info@gmail.com                                                            #
# Date:     March 2018                                                                      #
#                                                                                           #
#############################################################################################

from __future__ import absolute_import, division, print_function

__all__ = ['data', 'mdict', 'curve', 'image', 'video']


import os, sys
sys.path.append(os.path.dirname(__file__)[:-6][:-4][:-13])
#
from cognitivegeo.src.basic.data import data as data
from cognitivegeo.src.basic.matdict import matdict as mdict
from cognitivegeo.src.basic.curve import curve as curve
from cognitivegeo.src.basic.image import image as image
from cognitivegeo.src.basic.video import video as video


#############################################################################################
#                                                                                           #
# Author:   GeoPy Team                                                                      #
# Email:    geopy.info@gmail.com                                                            #
# Date:     March 2019                                                                      #
#                                                                                           #
#############################################################################################

# Create a window for import pre-stack seismic images


from PyQt5 import QtCore, QtGui, QtWidgets
import numpy as np
import sys, os
#
sys.path.append(os.path.dirname(__file__)[:-4][:-4][:-13])
from cognitivegeo.src.basic.data import data as basic_data
from cognitivegeo.src.psseismic.visualization import visualization as psseis_vis
from cognitivegeo.src.psseismic.analysis import analysis as psseis_ays
from cognitivegeo.src.vis.messager import messager as vis_msg

QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)
QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)


class importpsseisimageset(object):

    psseisdata = {}
    rootpath = ''
    #
    iconpath = os.path.dirname(__file__)
    dialog = None
    #
    imagelist = []

    def setupGUI(self, ImportPsSeisImageSet):
        ImportPsSeisImageSet.setObjectName("ImportPsSeisImageSet")
        ImportPsSeisImageSet.setFixedSize(400, 270)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(os.path.join(self.iconpath, "icons/image.png")),
                       QtGui.QIcon.Normal,
                       QtGui.QIcon.Off)
        ImportPsSeisImageSet.setWindowIcon(icon)
        self.lblimage = QtWidgets.QLabel(ImportPsSeisImageSet)
        self.lblimage.setObjectName("lblimage")
        self.lblimage.setGeometry(QtCore.QRect(10, 10, 110, 30))
        self.ldtimage = QtWidgets.QLineEdit(ImportPsSeisImageSet)
        self.ldtimage.setObjectName("ldtimage")
        self.ldtimage.setGeometry(QtCore.QRect(130, 10, 190, 30))
        self.btnimage = QtWidgets.QPushButton(ImportPsSeisImageSet)
        self.btnimage.setObjectName("btnimage")
        self.btnimage.setGeometry(QtCore.QRect(330, 10, 60, 30))
        self.lbltype = QtWidgets.QLabel(ImportPsSeisImageSet)
        self.lbltype.setObjectName("lbltype")
        self.lbltype.setGeometry(QtCore.QRect(10, 60, 110, 30))
        self.cbbtype = QtWidgets.QComboBox(ImportPsSeisImageSet)
        self.cbbtype.setObjectName("cbbtype")
        self.cbbtype.setGeometry(QtCore.QRect(130, 60, 260, 30))
        self.lbldims = QtWidgets.QLabel(ImportPsSeisImageSet)
        self.lbldims.setObjectName("lbldims")
        self.lbldims.setGeometry(QtCore.QRect(10, 110, 110, 30))
        self.ldtdimsinl = QtWidgets.QLineEdit(ImportPsSeisImageSet)
        self.ldtdimsinl.setObjectName("ldtdimsinl")
        self.ldtdimsinl.setGeometry(QtCore.QRect(130, 110, 60, 30))
        self.ldtdimsxl = QtWidgets.QLineEdit(ImportPsSeisImageSet)
        self.ldtdimsxl.setObjectName("ldtdimsxl")
        self.ldtdimsxl.setGeometry(QtCore.QRect(230, 110, 60, 30))
        self.ldtdimsz = QtWidgets.QLineEdit(ImportPsSeisImageSet)
        self.ldtdimsz.setObjectName("ldtdimsz")
        self.ldtdimsz.setGeometry(QtCore.QRect(330, 110, 60, 30))
        self.lblsave = QtWidgets.QLabel(ImportPsSeisImageSet)
        self.lblsave.setObjectName("lblsave")
        self.lblsave.setGeometry(QtCore.QRect(10, 160, 110, 30))
        self.ldtsave = QtWidgets.QLineEdit(ImportPsSeisImageSet)
        self.ldtsave.setObjectName("ldtsave")
        self.ldtsave.setGeometry(QtCore.QRect(130, 160, 130, 30))
        self.btnimportimage = QtWidgets.QPushButton(ImportPsSeisImageSet)
        self.btnimportimage.setObjectName("btnimportimage")
        self.btnimportimage.setGeometry(QtCore.QRect(120, 210, 160, 30))
        self.btnimportimage.setIcon(icon)
        #
        self.msgbox = QtWidgets.QMessageBox(ImportPsSeisImageSet)
        self.msgbox.setObjectName("msgbox")
        _center_x = ImportPsSeisImageSet.geometry().center().x()
        _center_y = ImportPsSeisImageSet.geometry().center().y()
        self.msgbox.setGeometry(QtCore.QRect(_center_x - 150, _center_y - 50, 300, 100))
        #
        self.retranslateGUI(ImportPsSeisImageSet)
        QtCore.QMetaObject.connectSlotsByName(ImportPsSeisImageSet)


    def retranslateGUI(self, ImportPsSeisImageSet):
        self.dialog = ImportPsSeisImageSet
        #
        _translate = QtCore.QCoreApplication.translate
        ImportPsSeisImageSet.setWindowTitle(_translate("ImportPsSeisImageSet", "Import Pre-stack Seismic ImageSet"))
        self.lblimage.setText(_translate("ImportPsSeisImageSet", "Select 2D images:"))
        self.lblimage.setAlignment(QtCore.Qt.AlignCenter)
        self.ldtimage.setText(_translate("ImportPsSeisImageSet", os.path.abspath(self.rootpath)))
        self.btnimage.setText(_translate("ImportPsSeisImageSet", "Browse"))
        self.btnimage.clicked.connect(self.clickBtnImage)
        self.lbltype.setText(_translate("ImportPsSeisImageSet", "       Orientation:"))
        self.lbltype.setAlignment(QtCore.Qt.AlignCenter)
        self.cbbtype.addItems(['Shot'])
        self.cbbtype.setItemIcon(0, QtGui.QIcon(os.path.join(self.iconpath, "icons/gather.png")))
        self.cbbtype.currentIndexChanged.connect(self.changeCbbType)
        self.lbldims.setText(_translate("ImportPsSeisImageSet", "Survey dimensions:"))
        self.lbldims.setAlignment(QtCore.Qt.AlignCenter)
        self.ldtdimsinl.setText('')
        self.ldtdimsinl.setAlignment(QtCore.Qt.AlignCenter)
        self.ldtdimsxl.setText('')
        self.ldtdimsxl.setAlignment(QtCore.Qt.AlignCenter)
        self.ldtdimsz.setText('')
        self.ldtdimsz.setAlignment(QtCore.Qt.AlignCenter)
        self.lblsave.setText(_translate("ImportPsSeisImageSet", "Output name:"))
        self.ldtsave.setText(_translate("ImportPsSeisImageSet", "image"))
        self.btnimportimage.setText(_translate("ImportPsSeisImageSet", "Import ImageSet"))
        self.btnimportimage.clicked.connect(self.clickBtnImportPsSeisImageSet)


    def clickBtnImage(self):
        _dialog = QtWidgets.QFileDialog()
        _file = _dialog.getOpenFileNames(None, 'Select Pre-stack Seismic Image(s)', self.rootpath,
                                         filter="Image files (*.jpg; *.png);; All files (*.*)")
        if len(_file[0]) > 0:
            self.imagelist = _file[0]
            self.ldtimage.setText(str(_file[0]))


    def changeCbbType(self):
        if self.cbbtype.currentIndex() == 0:
            self.ldtdimsinl.setEnabled(True)
            self.ldtdimsxl.setEnabled(True)
            self.ldtdimsz.setEnabled(True)


    def clickBtnImportPsSeisImageSet(self):
        self.refreshMsgBox()
        #
        _nimage = len(self.imagelist)
        if _nimage <= 0:
            vis_msg.print("ERROR in ImportPsSeisImageSet: No image selected for import", type='error')
            QtWidgets.QMessageBox.critical(self.msgbox,
                                           'Import Pre-stack Seismic ImageSet',
                                           'No image selected for import')
            return
        #
        _ninl = basic_data.str2int(self.ldtdimsinl.text())
        _nxl = basic_data.str2int(self.ldtdimsxl.text())
        _nz = basic_data.str2int(self.ldtdimsz.text())
        if _ninl is False or _nxl is False or _nz is False:
            vis_msg.print("ERROR in ImportPsSeisImageSet: Non-integer survey dimensions", type='error')
            QtWidgets.QMessageBox.critical(self.msgbox,
                                           'Import Pre-stack Seismic ImageSet',
                                           'Non-integer dimensions')
            return
        if _ninl <= 0 or _nxl <= 0 or _nz <= 0:
            vis_msg.print("ERROR in ImportPsSeisImageSet: Zero survey dimensions", type='error')
            QtWidgets.QMessageBox.critical(self.msgbox,
                                           'Import Pre-stack Seismic ImageSet',
                                           'Zero survey dimensions')
            return
        #
        # Progress dialog
        _pgsdlg = QtWidgets.QProgressDialog()
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(os.path.join(self.iconpath, "icons/image.png")),
                       QtGui.QIcon.Normal, QtGui.QIcon.Off)
        _pgsdlg.setWindowIcon(icon)
        _pgsdlg.setWindowTitle('Import Pre-stack Seismic ImageSet')
        _pgsdlg.setCancelButton(None)
        _pgsdlg.setWindowFlags(QtCore.Qt.WindowStaysOnTopHint)
        _pgsdlg.forceShow()
        _pgsdlg.setFixedWidth(400)
        #
        if self.cbbtype.currentIndex() == 0:
            _shots = np.linspace(0, len(self.imagelist)-1, len(self.imagelist), dtype=int)
            # _pgsdlg.setWindowTitle('Import ' + str(len(_shots)) + ' Pre-stack Seismic ImageSet')
            _imagedata = psseis_vis.loadPsSeisShot(self.imagelist, _shots,
                                                   ispref=False,
                                                   inlnum=_ninl, inlstart=0, inlstep=1,
                                                   xlnum=_nxl, xlstart=0, xlstep=1,
                                                   znum=_nz, zstart=0, zstep=-1,
                                                   qpgsdlg=_pgsdlg
                                                   )
        _psseisdata = {}
        if checkPsSeisData(_imagedata):
            _psseisdata[self.ldtsave.text()] = _imagedata
        #
        # add new data to psseisdata
        for key in _psseisdata.keys():
            if key in self.psseisdata.keys() and checkPsSeisData(self.psseisdata[key]):
                reply = QtWidgets.QMessageBox.question(self.msgbox, 'Import Pre-stack Seismic ImageSet',
                                                       key + ' already exists. Overwrite?',
                                                       QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                                                       QtWidgets.QMessageBox.No)
                if reply == QtWidgets.QMessageBox.No:
                    return
            self.psseisdata[key] = _psseisdata[key]
        #
        QtWidgets.QMessageBox.information(self.msgbox,
                                          "Import Pre-stack Seismic ImageSet",
                                          str(_nimage) + " image(s) imported as Pre-stack Seismic successfully")
        return


    def refreshMsgBox(self):
        _center_x = self.dialog.geometry().center().x()
        _center_y = self.dialog.geometry().center().y()
        self.msgbox.setGeometry(QtCore.QRect(_center_x - 150, _center_y - 50, 300, 100))


def checkPsSeisData(psseisdata):
    return psseis_ays.checkPsSeis(psseisdata)


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    ImportPsSeisImageSet = QtWidgets.QWidget()
    gui = importpsseisimageset()
    gui.setupGUI(ImportPsSeisImageSet)
    ImportPsSeisImageSet.show()
    sys.exit(app.exec_())


#############################################################################################
#                                                                                           #
# Author:   GeoPy Team                                                                      #
# Email:    geopy.info@gmail.com                                                            #
# Date:     December 2018                                                                   #
#                                                                                           #
#############################################################################################

# Create a window for exporting seismic survey


from PyQt5 import QtCore, QtGui, QtWidgets
import numpy as np
import sys, os
#
sys.path.append(os.path.dirname(__file__)[:-4][:-4][:-13])
from cognitivegeo.src.seismic.analysis import analysis as seis_ays
from cognitivegeo.src.vis.messager import messager as vis_msg

QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)
QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)


class exportsurvey(object):

    survinfo = {}
    rootpath = ''
    #
    iconpath = os.path.dirname(__file__)
    dialog = None

    def setupGUI(self, ExportSurvey):
        ExportSurvey.setObjectName("ExportSurvey")
        ExportSurvey.setFixedSize(500, 330)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(os.path.join(self.iconpath, "icons/survey.png")),
                       QtGui.QIcon.Normal,
                       QtGui.QIcon.Off)
        ExportSurvey.setWindowIcon(icon)
        self.lblsrvinfo = QtWidgets.QLabel(ExportSurvey)
        self.lblsrvinfo.setObjectName("lblsrvinfo")
        self.lblsrvinfo.setGeometry(QtCore.QRect(10, 10, 150, 30))
        self.lblstart = QtWidgets.QLabel(ExportSurvey)
        self.lblstart.setObjectName("lblstart")
        self.lblstart.setGeometry(QtCore.QRect(120, 50, 80, 30))
        self.lblend = QtWidgets.QLabel(ExportSurvey)
        self.lblend.setObjectName("lblend")
        self.lblend.setGeometry(QtCore.QRect(220, 50, 80, 30))
        self.lblstep = QtWidgets.QLabel(ExportSurvey)
        self.lblstep.setObjectName("lblstep")
        self.lblstep.setGeometry(QtCore.QRect(320, 50, 40, 30))
        self.lblnum = QtWidgets.QLabel(ExportSurvey)
        self.lblnum.setObjectName("lblnum")
        self.lblnum.setGeometry(QtCore.QRect(380, 50, 80, 30))
        self.lblinl = QtWidgets.QLabel(ExportSurvey)
        self.lblinl.setObjectName("lblinl")
        self.lblinl.setGeometry(QtCore.QRect(10, 90, 100, 30))
        self.lblxl = QtWidgets.QLabel(ExportSurvey)
        self.lblxl.setObjectName("lblxl")
        self.lblxl.setGeometry(QtCore.QRect(10, 130, 100, 30))
        self.lblz = QtWidgets.QLabel(ExportSurvey)
        self.lblz.setObjectName("lblz")
        self.lblz.setGeometry(QtCore.QRect(10, 170, 100, 30))
        self.ldtinlstart = QtWidgets.QLineEdit(ExportSurvey)
        self.ldtinlstart.setObjectName("ldtinlstart")
        self.ldtinlstart.setGeometry(QtCore.QRect(120, 90, 80, 30))
        self.ldtinlend = QtWidgets.QLineEdit(ExportSurvey)
        self.ldtinlend.setObjectName("ldtinlend")
        self.ldtinlend.setGeometry(QtCore.QRect(220, 90, 80, 30))
        self.ldtinlstep = QtWidgets.QLineEdit(ExportSurvey)
        self.ldtinlstep.setObjectName("ldtinlstep")
        self.ldtinlstep.setGeometry(QtCore.QRect(320, 90, 40, 30))
        self.ldtinlnum = QtWidgets.QLineEdit(ExportSurvey)
        self.ldtinlnum.setObjectName("ldtinlnum")
        self.ldtinlnum.setGeometry(QtCore.QRect(380, 90, 80, 30))
        self.ldtxlstart = QtWidgets.QLineEdit(ExportSurvey)
        self.ldtxlstart.setObjectName("ldtxlstart")
        self.ldtxlstart.setGeometry(QtCore.QRect(120, 130, 80, 30))
        self.ldtxlend = QtWidgets.QLineEdit(ExportSurvey)
        self.ldtxlend.setObjectName("ldtxlend")
        self.ldtxlend.setGeometry(QtCore.QRect(220, 130, 80, 30))
        self.ldtxlstep = QtWidgets.QLineEdit(ExportSurvey)
        self.ldtxlstep.setObjectName("ldtxlstep")
        self.ldtxlstep.setGeometry(QtCore.QRect(320, 130, 40, 30))
        self.ldtxlnum = QtWidgets.QLineEdit(ExportSurvey)
        self.ldtxlnum.setObjectName("ldtxlnum")
        self.ldtxlnum.setGeometry(QtCore.QRect(380, 130, 80, 30))
        self.ldtzstart = QtWidgets.QLineEdit(ExportSurvey)
        self.ldtzstart.setObjectName("ldtzstart")
        self.ldtzstart.setGeometry(QtCore.QRect(120, 170, 80, 30))
        self.ldtzend = QtWidgets.QLineEdit(ExportSurvey)
        self.ldtzend.setObjectName("ldtzend")
        self.ldtzend.setGeometry(QtCore.QRect(220, 170, 80, 30))
        self.ldtzstep = QtWidgets.QLineEdit(ExportSurvey)
        self.ldtzstep.setObjectName("ldtzlstep")
        self.ldtzstep.setGeometry(QtCore.QRect(320, 170, 40, 30))
        self.ldtznum = QtWidgets.QLineEdit(ExportSurvey)
        self.ldtznum.setObjectName("ldtznum")
        self.ldtznum.setGeometry(QtCore.QRect(380, 170, 80, 30))
        self.lblsave = QtWidgets.QLabel(ExportSurvey)
        self.lblsave.setObjectName("lblsave")
        self.lblsave.setGeometry(QtCore.QRect(10, 220, 100, 30))
        self.ldtsave = QtWidgets.QLineEdit(ExportSurvey)
        self.ldtsave.setObjectName("ldtsave")
        self.ldtsave.setGeometry(QtCore.QRect(120, 220, 240, 30))
        self.btnsave = QtWidgets.QPushButton(ExportSurvey)
        self.btnsave.setObjectName("btnsave")
        self.btnsave.setGeometry(QtCore.QRect(380, 220, 80, 30))
        self.btnexport = QtWidgets.QPushButton(ExportSurvey)
        self.btnexport.setObjectName("btnexport")
        self.btnexport.setGeometry(QtCore.QRect(170, 270, 160, 30))
        self.btnexport.setIcon(icon)
        #
        self.msgbox = QtWidgets.QMessageBox(ExportSurvey)
        self.msgbox.setObjectName("msgbox")
        _center_x = ExportSurvey.geometry().center().x()
        _center_y = ExportSurvey.geometry().center().y()
        self.msgbox.setGeometry(QtCore.QRect(_center_x - 150, _center_y - 50, 300, 100))
        #
        self.retranslateGUI(ExportSurvey)
        QtCore.QMetaObject.connectSlotsByName(ExportSurvey)


    def retranslateGUI(self, ExportSurvey):
        self.dialog = ExportSurvey
        #
        _translate = QtCore.QCoreApplication.translate
        ExportSurvey.setWindowTitle(_translate("ExportSurvey", "Export Survey"))
        self.lblsrvinfo.setText(_translate("ExportSurvey", "Survey information:"))
        self.lblstart.setText(_translate("ExportSurvey", "Start"))
        self.lblstart.setAlignment(QtCore.Qt.AlignCenter)
        self.lblend.setText(_translate("ExportSurvey", "End"))
        self.lblend.setAlignment(QtCore.Qt.AlignCenter)
        self.lblstep.setText(_translate("ExportSurvey", "Step"))
        self.lblstep.setAlignment(QtCore.Qt.AlignCenter)
        self.lblnum.setText(_translate("ExportSurvey", "Number"))
        self.lblnum.setAlignment(QtCore.Qt.AlignCenter)
        self.lblinl.setText(_translate("ExportSurvey", "Inline:"))
        self.lblinl.setAlignment(QtCore.Qt.AlignRight)
        self.lblxl.setText(_translate("ExportSurvey", "Crossline:"))
        self.lblxl.setAlignment(QtCore.Qt.AlignRight)
        self.lblz.setText(_translate("ExportSurvey", "Time/depth:"))
        self.lblz.setAlignment(QtCore.Qt.AlignRight)
        self.ldtinlstart.setEnabled(False)
        self.ldtinlstart.setAlignment(QtCore.Qt.AlignCenter)
        self.ldtinlend.setEnabled(False)
        self.ldtinlend.setAlignment(QtCore.Qt.AlignCenter)
        self.ldtinlstep.setEnabled(False)
        self.ldtinlstep.setAlignment(QtCore.Qt.AlignCenter)
        self.ldtinlnum.setEnabled(False)
        self.ldtinlnum.setAlignment(QtCore.Qt.AlignCenter)
        self.ldtxlstart.setEnabled(False)
        self.ldtxlstart.setAlignment(QtCore.Qt.AlignCenter)
        self.ldtxlend.setEnabled(False)
        self.ldtxlend.setAlignment(QtCore.Qt.AlignCenter)
        self.ldtxlstep.setEnabled(False)
        self.ldtxlstep.setAlignment(QtCore.Qt.AlignCenter)
        self.ldtxlnum.setEnabled(False)
        self.ldtxlnum.setAlignment(QtCore.Qt.AlignCenter)
        self.ldtzstart.setEnabled(False)
        self.ldtzstart.setAlignment(QtCore.Qt.AlignCenter)
        self.ldtzend.setEnabled(False)
        self.ldtzend.setAlignment(QtCore.Qt.AlignCenter)
        self.ldtzstep.setEnabled(False)
        self.ldtzstep.setAlignment(QtCore.Qt.AlignCenter)
        self.ldtznum.setEnabled(False)
        self.ldtznum.setAlignment(QtCore.Qt.AlignCenter)
        #
        self.lblsave.setText(_translate("ExportSurvey", "Save as:"))
        self.ldtsave.setText(_translate("ExportSurvey", ""))
        self.btnsave.setText(_translate("ExportSurvey", "Browse"))
        self.btnsave.clicked.connect(self.clickBtnSave)
        self.btnexport.setText(_translate("ExportSeisNpy", "Export Survey"))
        self.btnexport.clicked.connect(self.clickBtnExportSurvey)
        #
        self.refreshSeisInfo()


    def clickBtnSave(self):
        _dialog = QtWidgets.QFileDialog()
        _file = _dialog.getSaveFileName(None, 'Select Survey Numpy', self.rootpath,
                                        filter="Survey Numpy files (*.srv.npy);; All files (*.*)")
        if len(_file[0]) > 0:
            self.ldtsave.setText(_file[0])


    def clickBtnExportSurvey(self):
        self.refreshMsgBox()
        #
        if self.checkSurvInfo() is False:
            vis_msg.print("ERROR in ExportSurvey: No survey found", type='error')
            QtWidgets.QMessageBox.critical(self.msgbox,
                                           'Export Survey',
                                           'No survey found')
            return
        #
        if len(self.ldtsave.text()) < 1:
            vis_msg.print("ERROR in ExportSurvey: No name specified for survey", type='error')
            QtWidgets.QMessageBox.critical(self.msgbox,
                                           'Export Survey',
                                           'No name specified for survey')
            return
        #
        print("ExportSurvey: Export survey as %s" % self.ldtsave.text())
        #
        np.save(self.ldtsave.text(), self.survinfo)
        #
        QtWidgets.QMessageBox.information(self.msgbox,
                                          "Export Survey",
                                          "Survey exported successfully")
        return


    def refreshSeisInfo(self):
        if self.checkSurvInfo():
            _seisinfo = self.survinfo
            self.ldtinlstart.setText(str(_seisinfo['ILStart']))
            self.ldtinlend.setText(str(_seisinfo['ILEnd']))
            self.ldtinlstep.setText(str(_seisinfo['ILStep']))
            self.ldtinlnum.setText(str(_seisinfo['ILNum']))
            self.ldtxlstart.setText(str(_seisinfo['XLStart']))
            self.ldtxlend.setText(str(_seisinfo['XLEnd']))
            self.ldtxlstep.setText(str(_seisinfo['XLStep']))
            self.ldtxlnum.setText(str(_seisinfo['XLNum']))
            self.ldtzstart.setText(str(_seisinfo['ZStart']))
            self.ldtzend.setText(str(_seisinfo['ZEnd']))
            self.ldtzstep.setText(str(_seisinfo['ZStep']))
            self.ldtznum.setText(str(_seisinfo['ZNum']))


    def refreshMsgBox(self):
        _center_x = self.dialog.geometry().center().x()
        _center_y = self.dialog.geometry().center().y()
        self.msgbox.setGeometry(QtCore.QRect(_center_x - 150, _center_y - 50, 300, 100))


    def checkSurvInfo(self):
        self.refreshMsgBox()
        #
        return seis_ays.checkSeisInfo(self.survinfo)



if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    ExportSurvey = QtWidgets.QWidget()
    gui = exportsurvey()
    gui.setupGUI(ExportSurvey)
    ExportSurvey.show()
    sys.exit(app.exec_())


ASCII = 'ascii'
EBCDIC = 'cp037'

SUPPORTED_ENCODINGS = (ASCII, EBCDIC)


class UnsupportedEncodingError(Exception):

    def __init__(self, text, encoding):
        self._encoding = encoding
        super(UnsupportedEncodingError, self).__init__(text)

    @property
    def encoding(self):
        return self._encoding

    def __str__(self):
        return "{} not supported for encoding {}".format(self.args[0], self._encoding)

    def __repr__(self):
        return "{}({!r}, {!r}".format(self.__class__.__name__, self.args[0], self._encoding)


def is_supported_encoding(encoding):
    return encoding in SUPPORTED_ENCODINGS


COMMON_CHARS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789:_- '
COMMON_EBCDIC_CHARS = set(COMMON_CHARS.encode(EBCDIC))
COMMON_ASCII_CHARS = set(COMMON_CHARS.encode(ASCII))

def guess_encoding(bs, threshold=0.5):
    """Try to determine whether the encoding of byte stream b is an ASCII string or an EBCDIC string.

    Args:
        bs: A byte string (Python 2 - str; Python 3 - bytes)

    Returns:
        A string which can be used with the Python encoding functions: 'cp037' for EBCDIC, 'ascii' for ASCII or None
        if neither.
    """

    ebcdic_count = 0
    ascii_count = 0
    null_count = 0

    count = 0
    for b in bs:
        if b in COMMON_EBCDIC_CHARS:
            ebcdic_count +=1
        if b in COMMON_ASCII_CHARS:
            ascii_count +=1
        if b == 0:
            null_count += 1
        count += 1

    if count == 0:
        return None

    ebcdic_freq = ebcdic_count / count
    ascii_freq = ascii_count / count
    null_freq = null_count / count

    if null_freq == 1.0:
        return ASCII  # Doesn't matter

    if ebcdic_freq < threshold and ascii_freq >= threshold:
        return ASCII

    if ebcdic_freq >= threshold and ascii_freq < threshold:
        return EBCDIC

    if ebcdic_freq < threshold and ascii_freq < threshold:
        return None



    return None



from __future__ import print_function
import os, sys
sys.path.append(os.path.dirname(__file__)[:-6][:-4][:-13])
from cognitivegeo.src.segpy.encoding import ASCII
from cognitivegeo.src.segpy.packer import HeaderPacker

from cognitivegeo.src.segpy.trace_header import TraceHeaderRev1
from cognitivegeo.src.segpy.util import file_length, filename_from_handle
from cognitivegeo.src.segpy.datatypes import DATA_SAMPLE_FORMAT_TO_SEG_Y_TYPE, \
    SEG_Y_TYPE_DESCRIPTION, SEG_Y_TYPE_TO_CTYPE, size_in_bytes
from cognitivegeo.src.segpy.toolkit import (extract_revision,
                                            bytes_per_sample,
                                            read_binary_reel_header,
                                            read_trace_header,
                                            catalog_traces,
                                            read_binary_values,
                                            REEL_HEADER_NUM_BYTES,
                                            TRACE_HEADER_NUM_BYTES,
                                            read_textual_reel_header,
                                            read_extended_textual_headers,
                                            guess_textual_header_encoding)


def create_reader(fh, encoding=None, trace_header_format=TraceHeaderRev1, endian='>', progress=None):
    """Create a SegYReader (or one of its subclasses) based on performing
    a scan of SEG Y data.

    This function is the preferred method for creating SegYReader
    objects. It reads basic header information and attempts to build
    indexes for traces, CDP numbers (for 2D surveys), and inline and
    cross line co-ordinates (for 3D surveys) to facilitate subsequent
    random-access to traces.

    Args:
        fh: A file-like-object open in binary mode positioned such
            that the beginning of the reel header will be the next
            byte to be read. For disk-based SEG Y files, this is the
            beginning of the file.

        encoding: An optional text encoding for the textual headers. If
            None (the default) a heuristic will be used to guess the
            header encoding.

        trace_header_format: The class defining the layout of the trace header.

        endian: '>' for big-endian data (the standard and default), '<'
                for little-endian (non-standard)

        progress: A unary callable which will be passed a number
            between zero and one indicating the progress made. If
            provided, this callback will be invoked at least once with
            an argument equal to one.

    Raises:
        ValueError: ``fh`` is unsuitable for some reason, such as not
                    being open, not being seekable, not being in
                    binary mode, or being too short.

    Returns:
        A SegYReader object. Depending on the exact type of the
        SegYReader returned different capabilities may be
        available. Inspect the returned object to determine these
        capabilities, or be prepared for capabilities not defined in
        the SegYReader base class to be unavailable.  The underlying
        file-like object must remain open for the duration of use of
        the returned reader object. It is the caller's responsibility
        to close the underlying file.

    Example:

        with open('my_seismic_data.sgy', 'rb') as fh:
            reader = create_reader(fh)
            print(reader.num_traces())

    """
    if hasattr(fh, 'encoding') and fh.encoding is not None:
        raise TypeError(
            "SegYReader must be provided with a binary mode file object")

    if not fh.seekable():
        raise TypeError(
            "SegYReader must be provided with a seekable file object")

    if fh.closed:
        raise ValueError(
            "SegYReader must be provided with an open file object")

    num_file_bytes = file_length(fh)
    if num_file_bytes < REEL_HEADER_NUM_BYTES:
        raise ValueError(
            "SEG Y file {!r} of {} bytes is too short".format(
                filename_from_handle(fh),
                num_file_bytes))

    if endian not in ('<', '>'):
        raise ValueError("Unrecognised endian value {!r}".format(endian))

    if encoding is None:
        encoding = guess_textual_header_encoding(fh)

    if encoding is None:
        encoding = ASCII

    textual_reel_header = read_textual_reel_header(fh, encoding)
    binary_reel_header = read_binary_reel_header(fh, endian)
    extended_textual_header = read_extended_textual_headers(fh, binary_reel_header, encoding)
    revision = extract_revision(binary_reel_header)
    bps = bytes_per_sample(binary_reel_header, revision)

    trace_offset_catalog, trace_length_catalog, cdp_catalog, line_catalog = catalog_traces(fh, bps, trace_header_format,
                                                                                           endian, progress)

    # edited by Haibin Di on May 21 to resolve the issue of 3D of both cdp_catalog & line_catalog
    # if cdp_catalog is not None and line_catalog is None:
    #     return SegYReader2D(fh, textual_reel_header, binary_reel_header, extended_textual_header, trace_offset_catalog,
    #                         trace_length_catalog, cdp_catalog, trace_header_format, encoding, endian)
    #
    # if cdp_catalog is None and line_catalog is not None:
    #     return SegYReader3D(fh, textual_reel_header, binary_reel_header, extended_textual_header, trace_offset_catalog,
    #                         trace_length_catalog, line_catalog, trace_header_format, encoding, endian)
    #
    # return SegYReader(fh, textual_reel_header, binary_reel_header, extended_textual_header, trace_offset_catalog,
    #                   trace_length_catalog, trace_header_format, encoding, endian)

    if line_catalog is not None:
        return SegYReader3D(fh, textual_reel_header, binary_reel_header, extended_textual_header, trace_offset_catalog,
                            trace_length_catalog, line_catalog, trace_header_format, encoding, endian)
    elif cdp_catalog is not None:
        return SegYReader2D(fh, textual_reel_header, binary_reel_header, extended_textual_header, trace_offset_catalog,
                            trace_length_catalog, cdp_catalog, trace_header_format, encoding, endian)
    else:
        return SegYReader(fh, textual_reel_header, binary_reel_header, extended_textual_header, trace_offset_catalog,
                          trace_length_catalog, trace_header_format, encoding, endian)



class SegYReader(object):
    """A basic SEG Y reader.

    Use to obtain read the reel header, the trace_samples headers or trace_samples
    values. Traces can be accessed only by trace_samples index.
    """

    def __init__(self,
                 fh,
                 textual_reel_header,
                 binary_reel_header,
                 extended_textual_headers,
                 trace_offset_catalog,
                 trace_length_catalog,
                 trace_header_format,
                 encoding,
                 endian='>'):
        """Initialize a SegYReader around a file-like-object.

                Note:
            Usually a SegYReader is most easily constructed using the
            create_reader() function.

        Args:
            fh: A file-like object, which must support seeking and
            support binary reading.

            textual_reel_header: A sequence of forty 80-character Unicode strings
                containing header data.

            binary_reel_header: A dictionary containing reel header data.

            extended_textual_headers: A sequence of sequences of Unicode strings.

            trace_offset_catalog: A mapping from zero-based trace_samples index to
                the byte-offset to individual traces within the file.

            trace_length_catalog: A mapping from zero-based trace_samples index to the
                number of samples in that trace_samples.

            trace_header_format: The class defining the layout of the trace header.

            encoding: Either ASCII or EBCDIC.

            endian: '>' for big-endian data (the standard and default), '<' for
                little-endian (non-standard)

        """
        self._fh = fh
        self._endian = endian
        self._encoding = encoding

        self._textual_reel_header = textual_reel_header
        self._binary_reel_header = binary_reel_header
        self._extended_textual_headers = extended_textual_headers

        self._trace_header_packer = HeaderPacker(trace_header_format, endian)

        self._trace_offset_catalog = trace_offset_catalog
        self._trace_length_catalog = trace_length_catalog

        self._revision = extract_revision(self._binary_reel_header)
        self._bytes_per_sample = bytes_per_sample(
            self._binary_reel_header, self.revision)

    def trace_indexes(self):
        """An iterator over zero-based trace_samples indexes.

        Returns:
            An iterator which yields integers in the range zero to
            num_traces() - 1
        """
        return iter(self._trace_offset_catalog)

    def num_traces(self):
        """The number of traces"""
        return len(self._trace_offset_catalog)

    def max_num_trace_samples(self):
        """The number of samples in the trace_samples with the most samples."""
        return self._trace_length_catalog.value_max()

    def num_trace_samples(self, trace_index):
        """The number of samples in the specified trace_samples."""
        return self._trace_length_catalog[trace_index]

    def trace_samples(self, trace_index, start=None, stop=None):
        """Read a specific trace_samples.

        Args:
            trace_index: An integer in the range zero to num_traces() - 1

            start: Optional zero-based start sample index. The default
                is to read from the first (i.e. zeroth) sample.

            stop: Optional zero-based stop sample index. Following Python
                slice convention this is one beyond the end.

        Returns:
            A sequence of numeric trace_samples samples.

        Example:

            first_trace_samples = segy_reader.trace_samples(0)
            part_of_second_trace_samples = segy_reader.trace_samples(1, 1000, 2000)
        """
        if not (0 <= trace_index < self.num_traces()):
            raise ValueError("Trace index out of range.")

        num_samples_in_trace = self.num_trace_samples(trace_index)

        start_sample = start if start is not None else 0
        stop_sample = stop if stop is not None else num_samples_in_trace

        if not (0 <= stop_sample <= num_samples_in_trace):
            raise ValueError("trace_samples(): stop value {} out of range 0 to {}"
                             .format(stop, num_samples_in_trace))

        if not (0 <= start_sample <= stop_sample):
            raise ValueError("trace_samples(): start value {} out of range 0 to {}"
                             .format(start, stop_sample))

        dsf = self._binary_reel_header.data_sample_format
        seg_y_type = DATA_SAMPLE_FORMAT_TO_SEG_Y_TYPE[dsf]
        start_pos = (self._trace_offset_catalog[trace_index]
                     + TRACE_HEADER_NUM_BYTES
                     + start_sample * size_in_bytes(SEG_Y_TYPE_TO_CTYPE[seg_y_type]))
        num_samples_to_read = stop_sample - start_sample

        trace_values = read_binary_values(
            self._fh, start_pos, seg_y_type, num_samples_to_read, self._endian)
        return trace_values

    def trace_header(self, trace_index):
        """Read a specific trace_samples.

        Args:
            trace_index: An integer in the range zero to num_traces() - 1

        Returns:
            A TraceHeader corresponding to the requested trace_samples.

        Example:

            first_trace_header, first_trace_samples = segy_reader.trace_samples(0)
        """
        if not (0 <= trace_index < self.num_traces()):
            raise ValueError("Trace index {} out of range".format(trace_index))
        pos = self._trace_offset_catalog[trace_index]
        trace_header = read_trace_header(self._fh, self._trace_header_packer, pos)
        return trace_header

    @property
    def dimensionality(self):
        """The spatial dimensionality of the data.

        Returns:
            3 for 3D seismic volumes, 2 for 2D seismic lines, 1 for a
            single trace_samples, otherwise 0.

        """
        return self._dimensionality()

    def _dimensionality(self):
        return 1 if self.num_traces() == 1 else 0

    @property
    def textual_reel_header(self):
        """The textual real header.

        An immutable sequence of forty Unicode strings each 80 characters long.
        """
        return self._textual_reel_header

    @property
    def binary_reel_header(self):
        """The binary reel header.

        A dictionary containing data from the reel header.
        """
        return self._binary_reel_header

    @property
    def extended_textual_header(self):
        """A sequence of sequences of Unicode strings.

        If there were no headers, the sequence will be empty.
        """
        return self._extended_textual_headers


    @property
    def filename(self):
        """The filename.

        Returns:
            The filename if it could be determined, otherwise '<unknown>'
        """
        return filename_from_handle(self._fh)

    @property
    def revision(self):
        """The SEG Y revision.

        Returns:
            Either datatypes.SEGY_REVISION_0 or datatypes.SEGY_REVISION_1
        """
        return self._revision

    @property
    def bytes_per_sample(self):
        """The number of bytes per trace_samples sample.
        """
        return self._bytes_per_sample

    @property
    def data_sample_format(self):
        """The data type of the samples in machine-readable form.

        Returns:
            One of the values from datatypes.DATA_SAMPLE_FORMAT
        """
        return DATA_SAMPLE_FORMAT_TO_SEG_Y_TYPE[self._binary_reel_header.data_sample_format]

    @property
    def data_sample_format_description(self):
        """A descriptive human-readable description of the data sample format
        """
        return SEG_Y_TYPE_DESCRIPTION[self.data_sample_format]

    @property
    def encoding(self):
        """The encoding, of the data in the underlying file. Either ASCII ('ascii'),
           EBCDIC ('cp037') or None."""
        return self._encoding

    @property
    def endian(self):
        """The endianness of the data in the underlying file. Either '>' for big-endian or '<' for
           little endian or None."""
        return self._endian


class SegYReader3D(SegYReader):
    """A reader for 3D seismic data.

    In addition to the capabilities provided by the SegYReader base
    class, this reader provides an index to facilitate random access
    to individual traces via crossline and inline co-ordinates.
    """

    def __init__(self,
                 fh,
                 textual_reel_header,
                 binary_reel_header,
                 extended_textual_headers,
                 trace_offset_catalog,
                 trace_length_catalog,
                 line_catalog,
                 trace_header_format,
                 encoding,
                 endian='>'):
        """Initialize a SegYReader3D around a file-like-object.

                Note:
            Usually a SegYReader is most easily constructed using the
            create_reader() function.

        Args:
            fh: A file-like object, which must support seeking and
                support binary reading.

            binary_reel_header: A dictionary containing reel header data.

            trace_offset_catalog: A mapping from zero-based trace_samples indexes to
                the byte-offset to individual traces within the file.

            trace_length_catalog: A mapping from zero-based trace_samples indexes to
                the number of samples in that trace_samples.

            line_catalog: A mapping from (xline, inline) tuples to
                trace_indexes.

            trace_header_format: The class defining the layout of the trace header.

            encoding: Either ASCII or EBCDIC.

            endian: '>' for big-endian data (the standard and default), '<' for
                little-endian (non-standard)
        """
        super(SegYReader3D, self).__init__(fh, textual_reel_header, binary_reel_header, extended_textual_headers,
                                           trace_offset_catalog, trace_length_catalog, trace_header_format,
                                           encoding, endian)
        self._line_catalog = line_catalog
        self._num_inlines = None
        self._num_xlines = None

    def _dimensionality(self):
        return 3

    def inline_range(self):
        """A range encompassing inline numbers.

        The number of inlines within this range can be found with len(reader.inline_range()).

        Returns:
            A range() object with start set to the first inline number and stop set to
            one beyond the last inline number. The range always has a step of one, although
            this should not be taken as meaning that any intermediate inline number generated
            by the range is valid.
        """
        start = self._line_catalog.key_min()[0]
        stop = self._line_catalog.key_max()[0] + 1
        return range(start, stop)

    def num_inlines(self):
        """The number of distinct inlines in the survey.

        This number is not necessarily the same as the value returned by
        len(reader.inline_range()) as there may be missing inlines within the range.
        """
        if self._num_inlines is None:
            try:
                self._num_inlines = self._line_catalog.i_max - self._line_catalog.i_min + 1
            except AttributeError:
                self._num_inlines = len(set(i for i, j in self._line_catalog))
        return self._num_inlines

    def xline_range(self):
        """A range encompassing crossline numbers.

        The number of crosslines within this range can be found with len(reader.crossline_range()).

        Returns:
            A range() object with start set to the first crossline number and stop set to
            one beyond the last crossline number. The range always has a step of one, although
            this should not be taken as meaning that any intermediate crossline number generated
            by the range is valid.
        """
        start = self._line_catalog.key_min()[1]
        stop = self._line_catalog.key_max()[1] + 1
        return range(start, stop)

    def num_xlines(self):
        """The number of distinct crosslines in the survey.

        This number is not necessarily the same as the value returned by
        len(reader.xline_range()) as there may be missing crosslines within the range.
        """
        if self._num_xlines is None:
            try:
                self._num_xlines = self._line_catalog.j_max - self._line_catalog.j_min + 1
            except AttributeError:
                self._num_xlines = len(set(j for i, j in self._line_catalog))
        return self._num_xlines

    def inline_xline_numbers(self):
        """An iterator over all  (inline_number, xline_number) tuples
        corresponding to traces.
        """
        return iter(self._line_catalog)

    def has_trace_index(self, inline_xline):
        """Determine whether a specific trace_samples exists.

        Args:
            inline_xline: A 2-tuple of inline number, crossline number.

        Returns:
            True if the specified trace_samples exists, otherwise False.
        """
        return inline_xline in self._line_catalog

    def trace_index(self, inline_xline):
        """Obtain the trace_samples index given an xline and a inline.

        Note:
            Do not assume that all combinations of crossline and
            inline co-ordinates are valid.  The volume may not be
            rectangular.  Valid values can be obtained from the
            inline_xline_numbers() iterator.

            Furthermore, inline and crossline numbers should not be
            relied upon to be zero- or one-based indexes (although
            they may be).

        Args:
            inline_xline: A 2-tuple of inline number, crossline number.

        Returns:
            A trace_samples index which can be used with trace_samples().
        """
        return self._line_catalog[inline_xline]


class SegYReader2D(SegYReader):
    def __init__(self,
                 fh,
                 textual_reel_header,
                 binary_reel_header,
                 extended_textual_headers,
                 trace_offset_catalog,
                 trace_length_catalog,
                 cdp_catalog,
                 trace_header_format,
                 encoding,
                 endian='>'):
        """Initialize a SegYReader2D around a file-like-object.

        Note:
            Usually a SegYReader is most easily constructed using the
            create_reader() function.

        Args:
            fh: A file-like object, which must support seeking and
                support binary reading.

            binary_reel_header: A dictionary containing reel header data.

            trace_catalog_offset: A mapping from zero-based trace_samples index to
                the byte-offset to individual traces within the file.

            trace_length_catalog: A mapping from zero-based trace_samples indexes to
                the number of samples in that trace_samples.

            cdp_catalog: A mapping from CDP numbers to trace_indexes.

            trace_header_format: The class defining the layout of the trace header.

            encoding: Either ASCII or EBCDIC.

            endian: '>' for big-endian data (the standard and default), '<' for
                little-endian (non-standard)
        """
        super(SegYReader2D, self).__init__(fh, textual_reel_header, binary_reel_header, extended_textual_headers,
                                           trace_offset_catalog, trace_length_catalog, trace_header_format,
                                           encoding, endian)
        self._cdp_catalog = cdp_catalog

    def _dimensionality(self):
        return 2

    def cdp_numbers(self):
        """An iterator over all cdp numbers corresponding to traces.
        """
        return iter(self._cdp_catalog)

    def cdp_range(self):
        """A range encompassing CDP numbers.

        The number of CDPs within this range can be found with len(reader.cdp_range()).

        Returns:
            A range() object with start set to the first CDP number and stop set to
            one beyond the last CDP number. The range always has a step of one, although
            this should not be taken as meaning that any intermediate CDP number generated
            by the range is valid.
        """
        start = self._cdp_catalog.value_min()
        stop = self._cdp_catalog.value_max() + 1
        return range(start, stop)

    def num_cdps(self):
        """The number of distinct CDPs.

        This number is not necessarily the same as the value returned by
        len(reader.cdp_range()) as there may be missing CDPs.
        """
        return len(self._cdp_catalog)

    def has_trace_index(self, cdp_number):
        """Determine whether a specified trace_samples exists.

        Args:
            cdp_number: A CDP number.

        Returns:
            True if the trace_samples exists, otherwise False.
        """
        return self._cdp_catalog[cdp_number]

    def trace_index(self, cdp_number):
        """Obtain the trace_samples index given an xline and a inline.

        Args:
            cdp_number: A CDP number.

        Returns:
            A trace_samples index which can be used with trace_samples().
        """
        return self._cdp_catalog[cdp_number]


def main(argv=None):
    import sys

    if argv is None:
        argv = sys.argv[1:]

    class ProgressBar(object):

        def __init__(self, num_chars, character='.'):
            self._num_chars = num_chars
            self._character = character
            self._ratchet = 0

        def __call__(self, proportion):
            existing = self._num_marks(self._ratchet)
            required = self._num_marks(proportion)
            print(self._character * (required - existing), end='')
            self._ratchet = proportion

        def _num_marks(self, p):
            return int(round(p * self._num_chars))

    filename = argv[0]

    with open(filename, 'rb') as segy_file:
        segy_reader = create_reader(segy_file, progress=ProgressBar(30))
        print()
        print("Filename:             ", segy_reader.filename)
        print("SEG Y revision:       ", segy_reader.revision)
        print("Number of traces:     ", segy_reader.num_traces())
        print("Data format:          ",
              segy_reader.data_sample_format_description)
        print("Dimensionality:       ", segy_reader.dimensionality)

        try:
            print("Number of CDPs:       ", segy_reader.num_cdps())
        except AttributeError:
            pass

        try:
            print("Number of inlines:    ", segy_reader.num_inlines())
            print("Number of crosslines: ", segy_reader.num_xlines())
        except AttributeError:
            pass

        print("=== BEGIN TEXTUAL REEL HEADER ===")
        for line in segy_reader.textual_reel_header:
            print(line[3:])
        print("=== END TEXTUAL REEL HEADER ===")
        print()
        print("=== BEGIN EXTENDED TEXTUAL HEADER ===")
        print(segy_reader.extended_textual_header)
        print("=== END EXTENDED TEXTUAL_HEADER ===")


if __name__ == '__main__':
    main()



#############################################################################################
#                                                                                           #
# Author:   GeoPy Team                                                                      #
# Email:    geopy.info@gmail.com                                                            #
# Date:     January 2019                                                                    #
#                                                                                           #
#############################################################################################

# pre-stack seismic data visualization functions

from PyQt5 import QtCore
import sys, os
import numpy as np
from scipy import interpolate
import matplotlib.pyplot as plt
#
sys.path.append(os.path.dirname(__file__)[:-10][:-4][:-13])
from cognitivegeo.src.vis.font import font as vis_font
from cognitivegeo.src.vis.colormap import colormap as vis_cmap
from cognitivegeo.src.vis.messager import messager as vis_msg
from cognitivegeo.src.psseismic.analysis import analysis as psseis_ays


__all__ = ['visualization']


def remove_keymap_conflicts(new_keys_set):
    for prop in plt.rcParams:
        if prop.startswith('keymap.'):
            keys = plt.rcParams[prop]
            remove_list = set(keys) & new_keys_set
            for key in remove_list:
                keys.remove(key)


def first_shot(ax, shotlist, pref='', surf=''):
    """Go to the first shot."""
    volume = ax.volume
    ax.index = 0
    ax.set_title(pref + shotlist[ax.index] + surf)
    #
    nsample = np.shape(volume[shotlist[ax.index]]['ShotData'])[0]
    data = np.reshape(np.transpose(volume[shotlist[ax.index]]['ShotData'], [0, 2, 1]), [nsample, -1])
    ax.images[0].set_array(data)


def previous_shot(ax, shotlist, step=1, pref='', surf=''):
    """Go to the previous slice."""
    volume = ax.volume
    ax.index = (ax.index - step) % len(shotlist)  # wrap around using %
    ax.set_title(pref + shotlist[ax.index] + surf)
    #
    nsample = np.shape(volume[shotlist[ax.index]]['ShotData'])[0]
    data = np.reshape(np.transpose(volume[shotlist[ax.index]]['ShotData'], [0, 2, 1]), [nsample, -1])
    ax.images[0].set_array(data)


def next_shot(ax, shotlist, step=1, pref='', surf=''):
    """Go to the next slice."""
    volume = ax.volume
    ax.index = (ax.index + step) % len(shotlist)  # wrap around using %
    ax.set_title(pref + shotlist[ax.index] + surf)
    #
    nsample = np.shape(volume[shotlist[ax.index]]['ShotData'])[0]
    data = np.reshape(np.transpose(volume[shotlist[ax.index]]['ShotData'], [0, 2, 1]), [nsample, -1])
    ax.images[0].set_array(data)


def last_shot(ax, shotlist, pref='', surf=''):
    """Go to the last slice."""
    volume = ax.volume
    ax.index = -1
    ax.set_title(pref + shotlist[ax.index] + surf)
    #
    nsample = np.shape(volume[shotlist[ax.index]]['ShotData'])[0]
    data = np.reshape(np.transpose(volume[shotlist[ax.index]]['ShotData'], [0, 2, 1]), [nsample, -1])
    ax.images[0].set_array(data)


def plotPsSeisShot(psseis, shotlist=None,
                   colormap=None, flipcmap=False,
                   valuemin=-1.0, valuemax=1.0,
                   titlesurf='', colorbaron=False,
                   verbose=True):
    """
    Plot pre-stack shots (by matplotlib.pcolormesh), with each shot as a plot

    Args:
        psseis:     Pre-stach seismic as a dictionary
        shotlist:   List of the shots for visualization
        colormap:   Colormap for visualization as defined in vis/colormap.py. Default is None
        flipcmap:   Colormap flipping (true or false). Default is True
        valuemin:   Minimum value. Default is -1.0
        valuemax:   Maximum value. Default is 1.0
        titlesurf:  Surf of the plots. Default is ''
        colorbaron: Colorbar display  (true or false). Default is False
        verbose:    Verbose option for displaying the plotting message (true or false). Default is True

    Return:
        N/A
    """

    if psseis_ays.checkPsSeis(psseis) is False:
        vis_msg.print('ERROR in plotPsSeisShot: no pre-stack seismic found', type='error')
        sys.exit()
    #
    if shotlist is None:
        vis_msg.print('WARNING in plotPsSeisShot: plot all shots', type='warning')
        shotlist = list(psseis.keys())
    #
    nshot = len(shotlist)
    if verbose:
        print('Plot ' + str(nshot) + ' shots')
    for i in range(nshot):
        shot = shotlist[i]
        #
        if shot in psseis.keys():
            tracestart = psseis[shot]['ShotInfo']['XLStart']
            traceend = psseis[shot]['ShotInfo']['XLEnd']
            zstart = psseis[shot]['ShotInfo']['ZStart']
            zend = psseis[shot]['ShotInfo']['ZEnd']
            tracerange = psseis[shot]['ShotInfo']['XLRange']
            zrange = psseis[shot]['ShotInfo']['ZRange']
            x, y = np.meshgrid(tracerange, zrange)
            #
            seisdata = psseis[shot]['ShotData']
            plt.figure(facecolor='white')
            plt.pcolormesh(x, y, seisdata,
                           cmap=vis_cmap.makeColorMap(colormap, flipcmap),
                           shading='gouraud',
                           vmin=valuemin, vmax=valuemax)
            plt.xlim([tracestart, traceend])
            plt.ylim([zend, zstart])
            plt.title('Shot No.' + shot + titlesurf)
            plt.xlabel('Trace No.')
            plt.ylabel('Vertical (z) Depth/Time')
            if colorbaron:
                plt.colorbar()
    plt.show()
    #
    return True


def plotPsSeisShotPlayer(psseis, initshot=None,
                         colormap=None, flipcmap=False,
                         valuemin=-1.0, valuemax=1.0,
                         titlesurf='', colorbaron=False,
                         interpolation='bicubic',
                         playerconfig=None,
                         fontstyle=None,
                         qicon=None
                         ):
    """
    Plot pre-stack shots as a player (by matplotlib.pyplot.imshow)

    Args:
        psseis:         Pre-stach seismic as a dictionary
        initshot:       The shot for visualization when player starts
        colormap:       Colormap for visualization as defined in vis/colormap.py. Default is None
        flipcmap:       Colormap flipping (true or false). Default is True
        valuemin:       Minimum value. Default is -1.0
        valuemax:       Maximum value. Default is 1.0
        titlesurf:      Surf of the plots. Default is ''
        colorbaron:     Colorbar display  (true or false). Default is False
        interpolation:  Interpolation option. Default is 'bicubic'
        playerconfig:   Player configuration dictionary, with the following keys,
                            'First', 'Previsou', 'Backward', 'Pause',
                            'Forward', 'Next', 'Last', 'Interval'
        fontstyle:      Font style
        qicon:          QIcon for the visualization window

    Return:
        N/A
    """

    if psseis_ays.checkPsSeis(psseis) is False:
        vis_msg.print('ERROR in plotPsSeisShotPlayer: no pre-stack seismic found', type='error')
        sys.exit()
    #
    shotlist = list(sorted(psseis.keys()))
    #
    if initshot is None:
        vis_msg.print('WARNING in plotPsSeisShotPlayer: to be initialized by the first shot', type='warning')
        initshot = shotlist[0]
    if playerconfig is None:
        playerconfig = {}
        playerconfig['First'] = 'A'
        playerconfig['Previous'] = 'S'
        playerconfig['Backward'] = 'Z'
        playerconfig['Pause'] = 'X'
        playerconfig['Forward'] = 'C'
        playerconfig['Next'] = 'D'
        playerconfig['Last'] = 'F'
        playerconfig['Interval'] = 5
    #
    def process_key(event):
        fig = event.canvas.figure
        ax = fig.axes[0]
        if event.key == playerconfig['First'].lower():
            first_shot(ax, shotlist, pref='Shot No. ', surf=titlesurf)
            fig.canvas.draw()
        if event.key == playerconfig['Previous'].lower():
            previous_shot(ax, shotlist, step=playerconfig['Interval'], pref='Shot No. ', surf=titlesurf)
            fig.canvas.draw()
        if event.key == playerconfig['Next'].lower():
            next_shot(ax, shotlist, step=playerconfig['Interval'], pref='Shot No. ', surf=titlesurf)
            fig.canvas.draw()
        if event.key == playerconfig['Last'].lower():
            last_shot(ax, shotlist, pref='Shot No. ', surf=titlesurf)
            fig.canvas.draw()
        if event.key == playerconfig['Backward'].lower():
            while True:
                previous_shot(ax, shotlist, step=playerconfig['Interval'], pref='Shot No. ', surf=titlesurf)
                fig.canvas.draw()
                plt.pause(0.2)
        if event.key == playerconfig['Forward'].lower():
            while True:
                next_shot(ax, shotlist, step=playerconfig['Interval'], pref='Shot No. ', surf=titlesurf)
                fig.canvas.draw()
                plt.pause(0.2)
        if event.key == playerconfig['Pause'].lower():
            plt.pause(0)
    #
    remove_keymap_conflicts({playerconfig['First'].lower(), playerconfig['Previous'].lower(),
                             playerconfig['Backward'].lower(), playerconfig['Pause'].lower(),
                             playerconfig['Forward'].lower(), playerconfig['Next'].lower(),
                             playerconfig['Last'].lower()})
    #
    vis_font.updatePltFont(fontstyle)
    #
    fig, ax = plt.subplots(facecolor='white', figsize=(8, 8))
    ax.set_xlabel('Trace No.')
    ax.set_ylabel('Vertical (z) Depth/Time')
    volume = psseis
    ax.volume = volume
    ax.index = shotlist.index(initshot)
    ax.set_title('Shot No. ' + shotlist[ax.index] + titlesurf)
    #
    zstart = volume[shotlist[ax.index]]['ShotInfo']['ZStart']
    zend = volume[shotlist[ax.index]]['ShotInfo']['ZEnd']
    #
    ntrace = np.shape(volume[shotlist[ax.index]]['ShotData'])
    ntrace = ntrace[1] * ntrace[2]
    data = np.reshape(np.transpose(volume[shotlist[ax.index]]['ShotData'], [0, 2, 1]), [-1, ntrace])
    cat = ax.imshow(data,
                    cmap=vis_cmap.makeColorMap(colormap, flipcmap),
                    aspect='auto', #float(len(tracerange)) / float(len(zrange)),
                    extent=[1, ntrace, zend, zstart],
                    interpolation=interpolation,
                    vmin=valuemin, vmax=valuemax)

    #
    if colorbaron:
        fig.colorbar(cat)
    fig.canvas.mpl_connect('key_press_event', process_key)
    if qicon is not None:
        fig.canvas.set_window_title('2D Window - Pre-stack Gather')
        #
        # Commented by HD on June 7, 2018 to avoid crash
        # plt.get_current_fig_manager().window.setWindowIcon(qicon)
    plt.show()
    #
    return


def loadPsSeisShot(imagename, shots, ispref=True,
                   inlnum=100, xlnum=100, znum=100,
                   inlstart = 0, xlstart=0, zstart=0,
                   inlstep=1, xlstep=1, zstep=-1,
                   verbose=True, qpgsdlg=None):
    """
    Load pre-stack seismic shots from image files to dictionary

    Argus:
        imagename:  name of image files.
        shots:      list of shot in array [inl1, inl2, ...]
        ispref:     image name is given as pref
                    'Shot_XXX.jpg' is added with XXX representing shot No.
        inlnum:     number of inline slices fro create 3D matrix. Default is 100
        xlnum:      number of crossline slices for creating 3D matrix. Default is 100
        znum:       number of z slices for creating 3D matrix. Default is 100
        verbose:    flag for message display. Default is True

    Return:
        psseis:     pre-stack seismic in a dictionary, with two keys:
                        'ShotData' ----- 3D array of the data [Z/XL/IL]
                        'ShotInfo' ----- Shot information dictionary as defined in analysis.py

    Note:
        Negative z is used in the vertical direction
    """

    if np.ndim(shots) != 1:
        vis_msg.print('ERROR in loadPsSeisShot: 1D array of shot lists expected', type='error')
        sys.exit()

    xrange = np.linspace(0.0, 1.0, inlnum*xlnum)
    zrange = np.linspace(0.0, -1.0, znum)

    nshot = len(shots)
    if verbose:
        print('Load ' + str(nshot) + ' shot images to pre-stack seismic')

    psseis = {}

    if qpgsdlg is not None:
        qpgsdlg.setMaximum(nshot)

    for i in range(nshot):
        #
        if qpgsdlg is not None:
            QtCore.QCoreApplication.instance().processEvents()
            qpgsdlg.setValue(i)
        #
        if ispref:
            shotpath = imagename + 'Shot_' + str(shots[i]) + '.jpg'
        else:
            shotpath = imagename[i]
        #
        data = plt.imread(shotpath).astype(float)
        #
        image_data = 0.2989 * data[:, :, 0] + 0.5870 * data[:, :, 1] + 0.1140 * data[:, :, 2]
        if np.max(data[:, :, 0]) * np.max(data[:, :, 1]) * np.max(data[:, :, 2]) != 0:
            image_data = image_data * 3.0 / (np.max(data[:, :, 0]) + np.max(data[:, :, 1]) + np.max(data[:, :, 2]))
        if np.shape(data)[2] > 3:
            image_data = image_data + 1.0 - data[:, :, 3]
        #
        image_x = np.linspace(0.0, 1.0, np.shape(image_data)[1])
        image_z = np.linspace(-1.0, 0.0, np.shape(image_data)[0])
        f_interp = interpolate.interp2d(image_x, image_z, image_data)

        psdata = {}
        psdata['ShotData'] = np.transpose(np.reshape(f_interp(xrange, zrange), [znum, inlnum, xlnum]),
                                          [0, 2, 1])
        psdata['ShotInfo'] = psseis_ays.createShotInfo(psdata['ShotData'],
                                                       zstart=zstart, zstep=zstep,
                                                       xlstart=xlstart, xlstep=xlstep,
                                                       inlstart=inlstart, inlstep=inlstep)
        #
        psseis[str(shots[i])] = psdata
    #
    if qpgsdlg is not None:
        qpgsdlg.setValue(nshot)

    return psseis


class visualization:
    plotPsSeisShot = plotPsSeisShot
    plotPsSeisShotPlayer = plotPsSeisShotPlayer
    #
    loadPsSeisShot = loadPsSeisShot


#############################################################################################
#                                                                                           #
# Author:   GeoPy Team                                                                      #
# Email:    geopy.info@gmail.com                                                            #
# Date:     August 2018                                                                     #
#                                                                                           #
#############################################################################################

# Create a window for editing point sets


from PyQt5 import QtCore, QtGui, QtWidgets
import os, sys
import numpy as np
#
sys.path.append(os.path.dirname(__file__)[:-4][:-4][:-13])
from cognitivegeo.src.basic.data import data as basic_data
from cognitivegeo.src.basic.matdict import matdict as basic_mdt
from cognitivegeo.src.core.settings import settings as core_set
from cognitivegeo.src.pointset.analysis import analysis as point_ays
from cognitivegeo.src.vis.messager import messager as vis_msg
from cognitivegeo.src.gui.viewpointset import viewpointset as gui_viewpointset
from cognitivegeo.src.gui.editseispointset import editseispointset as gui_editseispointset
from cognitivegeo.src.gui.filterpointset import filterpointset as gui_filterpointset
from cognitivegeo.src.gui.plotvis2dpointsetcrossplt import plotvis2dpointsetcrossplt as gui_plotvis2dpointsetcrossplt

QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)
QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)


class editpointset(object):

    pointsetdata = {}
    rootpath = ''
    linestyle = core_set.Visual['Line']
    #
    iconpath = os.path.dirname(__file__)
    dialog = None


    def setupGUI(self, EditPointSet):
        EditPointSet.setObjectName("EditPointSet")
        EditPointSet.setFixedSize(300, 420)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(os.path.join(self.iconpath, "icons/point.png")),
                       QtGui.QIcon.Normal,
                       QtGui.QIcon.Off)
        EditPointSet.setWindowIcon(icon)
        #
        self.lblpoint = QtWidgets.QLabel(EditPointSet)
        self.lblpoint.setObjectName("lblpoint")
        self.lblpoint.setGeometry(QtCore.QRect(10, 10, 150, 30))
        self.lwgpoint = QtWidgets.QListWidget(EditPointSet)
        self.lwgpoint.setObjectName("lwgpoint")
        self.lwgpoint.setGeometry(QtCore.QRect(10, 50, 280, 200))
        self.lwgpoint.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.lblaction = QtWidgets.QLabel(EditPointSet)
        self.lblaction.setObjectName("lblaction")
        self.lblaction.setGeometry(QtCore.QRect(110, 270, 40, 30))
        self.cbbaction = QtWidgets.QComboBox(EditPointSet)
        self.cbbaction.setObjectName("cbbaction")
        self.cbbaction.setGeometry(QtCore.QRect(160, 270, 130, 30))
        self.lblrename = QtWidgets.QLabel(EditPointSet)
        self.lblrename.setObjectName("lblrename")
        self.lblrename.setGeometry(QtCore.QRect(160, 310, 40, 30))
        self.ldtrename = QtWidgets.QLineEdit(EditPointSet)
        self.ldtrename.setObjectName("ldtrename")
        self.ldtrename.setGeometry(QtCore.QRect(200, 310, 90, 30))
        self.cbbrename = QtWidgets.QComboBox(EditPointSet)
        self.cbbrename.setObjectName("cbbrename")
        self.cbbrename.setGeometry(QtCore.QRect(200, 310, 90, 30))
        #
        self.btnedit = QtWidgets.QPushButton(EditPointSet)
        self.btnedit.setObjectName("btnedit")
        self.btnedit.setGeometry(QtCore.QRect(100, 370, 100, 30))
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(os.path.join(self.iconpath, "icons/ok.png")),
                       QtGui.QIcon.Normal,
                       QtGui.QIcon.Off)
        self.btnedit.setIcon(icon)
        #
        self.msgbox = QtWidgets.QMessageBox(EditPointSet)
        self.msgbox.setObjectName("msgbox")
        _center_x = EditPointSet.geometry().center().x()
        _center_y = EditPointSet.geometry().center().y()
        self.msgbox.setGeometry(QtCore.QRect(_center_x - 150, _center_y - 50, 300, 100))
        #
        self.retranslateGUI(EditPointSet)
        QtCore.QMetaObject.connectSlotsByName(EditPointSet)


    def retranslateGUI(self, EditPointSet):
        self.dialog = EditPointSet
        #
        _translate = QtCore.QCoreApplication.translate
        EditPointSet.setWindowTitle(_translate("EditPointSet", "Edit PointSet"))
        self.lblpoint.setText(_translate("EditPointSet", "List of available pointsets:"))
        self.lblaction.setText(_translate("EditPointSet", "Action:"))
        self.cbbaction.addItems(['Copy', 'Delete', 'Rename',
                                 'Filter', 'Subset', 'Append',
                                 'Edit', 'View', 'Cross-plot'])
        self.cbbaction.setItemIcon(0, QtGui.QIcon(os.path.join(self.iconpath, "icons/copy.png")))
        self.cbbaction.setItemIcon(1, QtGui.QIcon(os.path.join(self.iconpath, "icons/delete.png")))
        self.cbbaction.setItemIcon(2, QtGui.QIcon(os.path.join(self.iconpath, "icons/rename.png")))
        self.cbbaction.setItemIcon(3, QtGui.QIcon(os.path.join(self.iconpath, "icons/filter.png")))
        self.cbbaction.setItemIcon(4, QtGui.QIcon(os.path.join(self.iconpath, "icons/export.png")))
        self.cbbaction.setItemIcon(5, QtGui.QIcon(os.path.join(self.iconpath, "icons/merge.png")))
        self.cbbaction.setItemIcon(6, QtGui.QIcon(os.path.join(self.iconpath, "icons/pen.png")))
        self.cbbaction.setItemIcon(7, QtGui.QIcon(os.path.join(self.iconpath, "icons/view.png")))
        self.cbbaction.setItemIcon(8, QtGui.QIcon(os.path.join(self.iconpath, "icons/plotpoint.png")))
        self.cbbaction.setCurrentIndex(7)
        self.cbbaction.currentIndexChanged.connect(self.changeCbbAction)
        self.lblrename.setText(_translate("EditPointSet", ""))
        self.lblrename.setVisible(False)
        self.ldtrename.setText(_translate("EditPointSet", ""))
        self.ldtrename.setVisible(False)
        self.cbbrename.setVisible(False)
        self.btnedit.setText(_translate("EditPointSet", "Apply"))
        self.btnedit.clicked.connect(self.clickBtnEditPointSet)
        #
        self.refreshLwgPoint()


    def changeCbbAction(self):
        if self.cbbaction.currentIndex() == 2:
            self.lblrename.setText('Name:')
            self.lblrename.setVisible(True)
            self.ldtrename.setVisible(True)
            self.cbbrename.setVisible(False)
        elif self.cbbaction.currentIndex() == 4:
            self.lblrename.setText('Size:')
            self.lblrename.setVisible(True)
            self.ldtrename.setVisible(True)
            self.cbbrename.setVisible(False)
        elif self.cbbaction.currentIndex() == 5:
            self.lblrename.setText('Select:')
            self.lblrename.setVisible(True)
            self.ldtrename.setVisible(True)
            self.cbbrename.setVisible(True)
        else:
            self.lblrename.setText('')
            self.lblrename.setVisible(False)
            self.ldtrename.setText('')
            self.ldtrename.setVisible(False)
            self.cbbrename.setVisible(False)


    def clickBtnEditPointSet(self):
        self.refreshMsgBox()
        #
        _pointlist = self.lwgpoint.selectedItems()
        if len(_pointlist) < 1:
            vis_msg.print("ERROR in EditPointSet: No pointset selected for editing", type='error')
            QtWidgets.QMessageBox.critical(self.msgbox,
                                           'Edit PointSet',
                                           'No pointset selected for editing')
            return
        #
        if self.cbbaction.currentIndex() == 0:
            self.pointsetdata[_pointlist[0].text()+'_copy'] = self.pointsetdata[_pointlist[0].text()].copy()
        if self.cbbaction.currentIndex() == 1:
            self.pointsetdata.pop(_pointlist[0].text())
        if self.cbbaction.currentIndex() == 2:
            if len(self.ldtrename.text()) < 1:
                vis_msg.print("ERROR in EditPointSet: No name specified for rename", type='error')
                QtWidgets.QMessageBox.critical(self.msgbox,
                                               'Edit PointSet',
                                               'No name specified for rename')
                return
            if self.ldtrename.text() in self.pointsetdata.keys():
                reply = QtWidgets.QMessageBox.question(self.msgbox, 'Edit PointSet',
                                                       self.ldtrename.text() + ' already exists. Overwrite?',
                                                       QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                                                       QtWidgets.QMessageBox.No)

                if reply == QtWidgets.QMessageBox.No:
                    return
            self.pointsetdata[self.ldtrename.text()] = self.pointsetdata[_pointlist[0].text()]
            self.pointsetdata.pop(_pointlist[0].text())
        if self.cbbaction.currentIndex() == 3:
            _filter = QtWidgets.QDialog()
            _gui = gui_filterpointset()
            _gui.pointsetname = _pointlist[0].text()
            _gui.pointsetdata = self.pointsetdata[_pointlist[0].text()].copy()
            _gui.setupGUI(_filter)
            _filter.exec()
            self.pointsetdata[_pointlist[0].text()] = _gui.pointsetdata.copy()
            _filter.show()
        if self.cbbaction.currentIndex() == 4:
            _size = basic_data.str2int(self.ldtrename.text())
            if _size is False:
                vis_msg.print("ERROR in EditPointSet: Non-integer size", type='error')
                QtWidgets.QMessageBox.critical(self.msgbox,
                                               'Edit PointSet',
                                               'Non-integer size')
                return
            self.pointsetdata[_pointlist[0].text()] = basic_mdt.retrieveDictRandom(self.pointsetdata[_pointlist[0].text()],
                                                                                batch_size=_size)
        if self.cbbaction.currentIndex() == 5:
            if self.cbbrename.currentText() not in self.pointsetdata.keys():
                vis_msg.print("ERROR in EditPointSet: No pointset selected for appending", type='error')
                QtWidgets.QMessageBox.critical(self.msgbox,
                                               'Edit PointSet',
                                               'No pointset selected for appending')
                return
            _pts = {}
            for f in self.pointsetdata[self.cbbrename.currentText()].keys():
                if f in self.pointsetdata[_pointlist[0].text()].keys():
                    _pts[f] = np.concatenate((self.pointsetdata[_pointlist[0].text()][f],
                                              self.pointsetdata[self.cbbrename.currentText()][f]), axis=0)
            self.pointsetdata[_pointlist[0].text()] = _pts
        if self.cbbaction.currentIndex() == 6:
            _editpoint = QtWidgets.QDialog()
            _gui = gui_editseispointset()
            _gui.seispointsetdata = self.pointsetdata[_pointlist[0].text()]
            _gui.rootpath = self.rootpath
            _gui.setupGUI(_editpoint)
            _editpoint.exec()
            self.pointsetdata[_pointlist[0].text()] = _gui.seispointsetdata
            _editpoint.show()
        if self.cbbaction.currentIndex() == 7:
            _view = QtWidgets.QDialog()
            _gui = gui_viewpointset()
            _gui.pointsetname = _pointlist[0].text()
            _gui.pointsetdata = self.pointsetdata[_pointlist[0].text()]
            _gui.linestyle = self.linestyle
            _gui.setupGUI(_view)
            _view.exec()
            # self.pointsetdata = _gui.pointsetdata.copy()
            _view.show()
        if self.cbbaction.currentIndex() == 8:
            _cplt = QtWidgets.QDialog()
            _gui = gui_plotvis2dpointsetcrossplt()
            _gui.pointsetdata = {}
            _gui.pointsetdata[_pointlist[0].text()] = self.pointsetdata[_pointlist[0].text()]
            _gui.linestyle = self.linestyle
            _gui.setupGUI(_cplt)
            _cplt.exec()
            _cplt.show()
        #
        self.refreshLwgPoint()
        # if self.cbbaction.currentIndex() != 3:
            # QtWidgets.QMessageBox.information(self.msgbox,
            #                                   "Edit PointSet",
            #                                   "PointSet edited successfully")
        return


    def refreshLwgPoint(self):
        self.lwgpoint.clear()
        self.cbbrename.clear()
        if len(self.pointsetdata.keys()) > 0:
            for i in sorted(self.pointsetdata.keys()):
                if self.checkPointSet(i):
                    item = QtWidgets.QListWidgetItem(self.lwgpoint)
                    item.setText(i)
                    self.lwgpoint.addItem(item)
            #
            self.cbbrename.addItems(sorted(self.pointsetdata.keys()))


    def refreshMsgBox(self):
        _center_x = self.dialog.geometry().center().x()
        _center_y = self.dialog.geometry().center().y()
        self.msgbox.setGeometry(QtCore.QRect(_center_x - 150, _center_y - 50, 300, 100))


    def checkPointSet(self, name):
        return point_ays.checkPoint(self.pointsetdata[name])


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    EditPointSet = QtWidgets.QWidget()
    gui = editpointset()
    gui.setupGUI(EditPointSet)
    EditPointSet.show()
    sys.exit(app.exec_())